import os
import shutil
import unittest
import pytest
from aiohttp.web import HTTPBadRequest, HTTPNotFound

from ...admin.request_context import AdminRequestContext
from ...ledger.base import BaseLedger
from ...ledger.multiple_ledger.base_manager import BaseMultipleLedgerManager
from ...protocols.endorse_transaction.v1_0.manager import (
    TransactionManager,
    TransactionManagerError,
)
from ...protocols.endorse_transaction.v1_0.models.transaction_record import (
    TransactionRecord,
)
from ...revocation.error import RevocationError
from ...storage.askar import AskarStorage
from ...storage.base import BaseStorage
from ...storage.error import StorageError
from ...tests import mock
from ...utils.testing import create_test_profile
from .. import routes as test_module
from ..manager import RevocationManager
from ..models.issuer_rev_reg_record import IssuerRevRegRecord
from multidict import MultiDict
import json


class TestRevocationRoutes(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.context = AdminRequestContext.test_context({}, self.profile)
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

        self.test_did = "sample-did"

        self.author_profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "author-key",
            }
        )
        self.author_profile.settings.set_value("endorser.author", True)
        self.author_context = AdminRequestContext.test_context({}, self.author_profile)
        self.author_request_dict = {
            "context": self.author_context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.author_request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.author_request_dict[k],
            headers={"x-api-key": "author-key"},
        )

    async def test_validate_cred_rev_rec_qs_and_revoke_req(self):
        for req in (
            test_module.CredRevRecordQueryStringSchema(),
            test_module.RevokeRequestSchema(),
        ):
            req.validate_fields(
                {
                    "rev_reg_id": test_module.INDY_REV_REG_ID_EXAMPLE,
                    "cred_rev_id": test_module.INDY_CRED_REV_ID_EXAMPLE,
                }
            )
            req.validate_fields({"cred_ex_id": test_module.UUID4_EXAMPLE})
            with self.assertRaises(test_module.ValidationError):
                req.validate_fields({})
            with self.assertRaises(test_module.ValidationError):
                req.validate_fields({"rev_reg_id": test_module.INDY_REV_REG_ID_EXAMPLE})
            with self.assertRaises(test_module.ValidationError):
                req.validate_fields({"cred_rev_id": test_module.INDY_CRED_REV_ID_EXAMPLE})
            with self.assertRaises(test_module.ValidationError):
                req.validate_fields(
                    {
                        "rev_reg_id": test_module.INDY_REV_REG_ID_EXAMPLE,
                        "cred_ex_id": test_module.UUID4_EXAMPLE,
                    }
                )
            with self.assertRaises(test_module.ValidationError):
                req.validate_fields(
                    {
                        "cred_rev_id": test_module.INDY_CRED_REV_ID_EXAMPLE,
                        "cred_ex_id": test_module.UUID4_EXAMPLE,
                    }
                )
            with self.assertRaises(test_module.ValidationError):
                req.validate_fields(
                    {
                        "rev_reg_id": test_module.INDY_REV_REG_ID_EXAMPLE,
                        "cred_rev_id": test_module.INDY_CRED_REV_ID_EXAMPLE,
                        "cred_ex_id": test_module.UUID4_EXAMPLE,
                    }
                )

    async def test_revoke(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "rev_reg_id": "rr_id",
                "cred_rev_id": "23",
                "publish": "false",
            }
        )

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_mgr.return_value.revoke_credential = mock.CoroutineMock()

            await test_module.revoke(self.request)

            mock_response.assert_called_once_with({})

    async def test_revoke_endorser_no_conn_id_by_cred_ex_id(self):
        self.author_request.json = mock.CoroutineMock(
            return_value={
                "rev_reg_id": "rr_id",
                "cred_rev_id": "23",
                "publish": "false",
            }
        )

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(
                test_module,
                "get_endorser_connection_id",
                mock.CoroutineMock(return_value="dummy-conn-id"),
            ),
            mock.patch.object(test_module.web, "json_response"),
            mock.patch.object(TransactionManager, "create_record", mock.CoroutineMock()),
        ):
            mock_mgr.return_value.revoke_credential = mock.CoroutineMock(
                return_value={"result": "..."}
            )

            await test_module.revoke(self.author_request)

    async def test_revoke_endorser_by_cred_ex_id(self):
        self.author_request.json = mock.CoroutineMock(
            return_value={
                "rev_reg_id": "rr_id",
                "cred_rev_id": "23",
                "publish": "false",
                "connection_id": "dummy-conn-id",
                "cred_ex_id": "dummy-cxid",
            }
        )

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(test_module.web, "json_response"),
            mock.patch.object(
                test_module,
                "get_endorser_connection_id",
                mock.CoroutineMock(return_value="test_conn_id"),
            ),
            mock.patch.object(TransactionManager, "create_record", mock.CoroutineMock()),
        ):
            mock_mgr.return_value.revoke_credential = mock.CoroutineMock(
                return_value={"result": "..."}
            )

            await test_module.revoke(self.author_request)

    async def test_revoke_endorser_no_conn_id(self):
        self.author_request.json = mock.CoroutineMock(
            return_value={
                "rev_reg_id": "rr_id",
                "cred_rev_id": "23",
                "publish": "false",
            }
        )

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(
                test_module,
                "get_endorser_connection_id",
                mock.CoroutineMock(return_value=None),
            ),
            mock.patch.object(test_module.web, "json_response"),
            mock.patch.object(TransactionManager, "create_record", mock.CoroutineMock()),
        ):
            mock_mgr.return_value.revoke_credential = mock.CoroutineMock(
                return_value={"result": "..."}
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.revoke(self.author_request)

    async def test_revoke_endorser(self):
        self.author_request.json = mock.CoroutineMock(
            return_value={
                "rev_reg_id": "rr_id",
                "cred_rev_id": "23",
                "publish": "false",
                "connection_id": "dummy-conn-id",
            }
        )

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(test_module.web, "json_response"),
            mock.patch.object(
                test_module,
                "get_endorser_connection_id",
                mock.CoroutineMock(return_value="test_conn_id"),
            ),
            mock.patch.object(TransactionManager, "create_record", mock.CoroutineMock()),
        ):
            mock_mgr.return_value.revoke_credential = mock.CoroutineMock(
                return_value={"result": "..."}
            )

            await test_module.revoke(self.author_request)

    async def test_revoke_endorser_x(self):
        self.author_request.json = mock.CoroutineMock(
            return_value={
                "rev_reg_id": "rr_id",
                "cred_rev_id": "23",
                "publish": "false",
            }
        )

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(
                test_module,
                "get_endorser_connection_id",
                mock.CoroutineMock(return_value=None),
            ),
        ):
            mock_mgr.return_value.revoke_credential = mock.CoroutineMock()
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.revoke(self.author_request)

    async def test_revoke_by_cred_ex_id(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "cred_ex_id": "dummy-cxid",
                "publish": "false",
            }
        )

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_mgr.return_value.revoke_credential = mock.CoroutineMock()

            await test_module.revoke(self.request)

            mock_response.assert_called_once_with({})

    async def test_revoke_not_found(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "rev_reg_id": "rr_id",
                "cred_rev_id": "23",
                "publish": "false",
            }
        )

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_mgr.return_value.revoke_credential = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.revoke(self.request)

    async def test_publish_revocations(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            pub_pending = mock.CoroutineMock()
            mock_mgr.return_value.publish_pending_revocations = mock.CoroutineMock(
                return_value=({}, pub_pending.return_value)
            )

            await test_module.publish_revocations(self.request)

            mock_response.assert_called_once_with({"rrid2crid": pub_pending.return_value})

    async def test_publish_revocations_x(self):
        self.request.json = mock.CoroutineMock()

        with mock.patch.object(
            test_module, "RevocationManager", autospec=True
        ) as mock_mgr:
            pub_pending = mock.CoroutineMock(side_effect=test_module.RevocationError())
            mock_mgr.return_value.publish_pending_revocations = pub_pending

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.publish_revocations(self.request)

    async def test_publish_revocations_endorser(self):
        self.author_request.json = mock.CoroutineMock(return_value={})

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(
                test_module,
                "get_endorser_connection_id",
                mock.CoroutineMock(return_value="dummy-conn-id"),
            ),
            mock.patch.object(
                TransactionManager,
                "create_record",
                mock.CoroutineMock(return_value=TransactionRecord()),
            ),
            mock.patch.object(
                TransactionManager,
                "create_request",
                mock.CoroutineMock(),
            ),
        ):
            pub_pending = mock.CoroutineMock()
            mock_mgr.return_value.publish_pending_revocations = mock.CoroutineMock(
                return_value=(
                    [
                        {"result": "..."},
                        {"result": "..."},
                    ],
                    pub_pending.return_value,
                )
            )

            result = await test_module.publish_revocations(self.author_request)
            assert result.status == 200

            # Auto endorsement
            self.author_request_dict["context"].settings["endorser.auto_request"] = True
            self.author_request = mock.MagicMock(
                app={},
                match_info={},
                query={},
                __getitem__=lambda _, k: self.author_request_dict[k],
                headers={"x-api-key": "author-key"},
            )
            self.author_request.json = mock.CoroutineMock()
            result = await test_module.publish_revocations(self.author_request)
            assert result.status == 200

    async def test_publish_revocations_endorser_exceptions(self):
        self.author_request.json = mock.CoroutineMock(return_value={})
        with mock.patch.object(
            test_module, "TransactionManager", autospec=True
        ) as mock_txn_mgr:
            mock_txn_mgr.return_value.create_record = mock.CoroutineMock(
                side_effect=StorageError()
            )
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.publish_revocations(self.author_request)

        with mock.patch.object(
            test_module, "TransactionManager", autospec=True
        ) as mock_txn_mgr:
            mock_txn_mgr.return_value.create_request = mock.CoroutineMock(
                side_effect=[StorageError(), TransactionManagerError()]
            )
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.publish_revocations(self.author_request)
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.publish_revocations(self.author_request)

    async def test_publish_revocations_endorser_x(self):
        self.author_request.json = mock.CoroutineMock()

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(
                test_module,
                "get_endorser_connection_id",
                mock.CoroutineMock(return_value=None),
            ),
            mock.patch.object(test_module.web, "json_response"),
        ):
            pub_pending = mock.CoroutineMock()
            mock_mgr.return_value.publish_pending_revocations = mock.CoroutineMock(
                return_value=(
                    [
                        {"result": "..."},
                        {"result": "..."},
                    ],
                    pub_pending.return_value,
                )
            )
            pub_pending = mock.CoroutineMock()
            mock_mgr.return_value.publish_pending_revocations = pub_pending
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.publish_revocations(self.author_request)

    async def test_clear_pending_revocations(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            clear_pending = mock.CoroutineMock()
            mock_mgr.return_value.clear_pending_revocations = clear_pending

            await test_module.clear_pending_revocations(self.request)

            mock_response.assert_called_once_with(
                {"rrid2crid": clear_pending.return_value}
            )

    async def test_clear_pending_revocations_x(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(
                test_module, "RevocationManager", autospec=True
            ) as mock_mgr,
            mock.patch.object(test_module.web, "json_response"),
        ):
            clear_pending = mock.CoroutineMock(side_effect=test_module.StorageError())
            mock_mgr.return_value.clear_pending_revocations = clear_pending

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.clear_pending_revocations(self.request)

    async def test_create_rev_reg(self):
        CRED_DEF_ID = f"{self.test_did}:3:CL:1234:default"
        self.request.json = mock.CoroutineMock(
            return_value={
                "max_cred_num": "1000",
                "credential_definition_id": CRED_DEF_ID,
            }
        )

        with (
            mock.patch.object(AskarStorage, "find_all_records"),
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_indy_revoc.return_value = mock.MagicMock(
                init_issuer_registry=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        generate_registry=mock.CoroutineMock(),
                        serialize=mock.MagicMock(return_value="dummy"),
                    )
                )
            )

            result = await test_module.create_rev_reg(self.request)
            mock_json_response.assert_called_once_with({"result": "dummy"})
            assert result is mock_json_response.return_value

    async def test_create_rev_reg_no_such_cred_def(self):
        CRED_DEF_ID = f"{self.test_did}:3:CL:1234:default"
        self.request.json = mock.CoroutineMock(
            return_value={
                "max_cred_num": "1000",
                "credential_definition_id": CRED_DEF_ID,
            }
        )

        with (
            mock.patch.object(
                BaseStorage, "find_all_records", autospec=True
            ) as mock_find,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_find.return_value = False

            with self.assertRaises(HTTPNotFound):
                await test_module.create_rev_reg(self.request)
            mock_json_response.assert_not_called()

    async def test_create_rev_reg_no_revo_support(self):
        CRED_DEF_ID = f"{self.test_did}:3:CL:1234:default"
        self.request.json = mock.CoroutineMock(
            return_value={
                "max_cred_num": "1000",
                "credential_definition_id": CRED_DEF_ID,
            }
        )

        with (
            mock.patch.object(
                AskarStorage, "find_all_records", autospec=True
            ) as mock_find,
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_find.return_value = True
            mock_indy_revoc.return_value = mock.MagicMock(
                init_issuer_registry=mock.CoroutineMock(
                    side_effect=test_module.RevocationNotSupportedError(
                        error_code="dummy"
                    )
                )
            )

            with self.assertRaises(HTTPBadRequest):
                await test_module.create_rev_reg(self.request)

            mock_json_response.assert_not_called()

    async def test_rev_regs_created(self):
        CRED_DEF_ID = f"{self.test_did}:3:CL:1234:default"
        self.request.query = {
            "cred_def_id": CRED_DEF_ID,
            "state": test_module.IssuerRevRegRecord.STATE_ACTIVE,
        }

        with (
            mock.patch.object(
                test_module.IssuerRevRegRecord, "query", mock.CoroutineMock()
            ) as mock_query,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_query.return_value = [mock.MagicMock(revoc_reg_id="dummy")]

            result = await test_module.rev_regs_created(self.request)
            mock_json_response.assert_called_once_with({"rev_reg_ids": ["dummy"]})
            assert result is mock_json_response.return_value

    async def test_get_rev_reg(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with (
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        serialize=mock.MagicMock(return_value="dummy")
                    )
                )
            )

            result = await test_module.get_rev_reg(self.request)
            mock_json_response.assert_called_once_with({"result": "dummy"})
            assert result is mock_json_response.return_value

    async def test_get_rev_reg_not_found(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with (
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    side_effect=test_module.StorageNotFoundError(error_code="dummy")
                )
            )

            with self.assertRaises(HTTPNotFound):
                await test_module.get_rev_reg(self.request)
            mock_json_response.assert_not_called()

    async def test_get_rev_reg_issued(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with (
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "retrieve_by_revoc_reg_id",
                mock.CoroutineMock(),
            ),
            mock.patch.object(
                test_module.IssuerCredRevRecord,
                "query_by_ids",
                mock.CoroutineMock(),
            ) as mock_query,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_query.return_value = [{"...": "..."}, {"...": "..."}]
            result = await test_module.get_rev_reg_issued_count(self.request)

            mock_json_response.assert_called_once_with({"result": 2})
            assert result is mock_json_response.return_value

    async def test_get_rev_reg_issued_x(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with mock.patch.object(
            test_module.IssuerRevRegRecord,
            "retrieve_by_revoc_reg_id",
            mock.CoroutineMock(),
        ) as mock_retrieve:
            mock_retrieve.side_effect = test_module.StorageNotFoundError("no such rec")

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.get_rev_reg_issued(self.request)

    async def test_get_cred_rev_record(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        CRED_REV_ID = "1"

        self.request.query = {
            "rev_reg_id": REV_REG_ID,
            "cred_rev_id": CRED_REV_ID,
        }

        with (
            mock.patch.object(
                test_module.IssuerCredRevRecord,
                "retrieve_by_ids",
                mock.CoroutineMock(),
            ) as mock_retrieve,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value="dummy")
            )
            result = await test_module.get_cred_rev_record(self.request)

            mock_json_response.assert_called_once_with({"result": "dummy"})
            assert result is mock_json_response.return_value

    async def test_get_cred_rev_record_by_cred_ex_id(self):
        CRED_EX_ID = test_module.UUID4_EXAMPLE

        self.request.query = {"cred_ex_id": CRED_EX_ID}

        with (
            mock.patch.object(
                test_module.IssuerCredRevRecord,
                "retrieve_by_cred_ex_id",
                mock.CoroutineMock(),
            ) as mock_retrieve,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value="dummy")
            )
            result = await test_module.get_cred_rev_record(self.request)

            mock_json_response.assert_called_once_with({"result": "dummy"})
            assert result is mock_json_response.return_value

    async def test_get_cred_rev_record_not_found(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        CRED_REV_ID = "1"

        self.request.json = mock.CoroutineMock(
            return_value={
                "rev_reg_id": REV_REG_ID,
                "cred_rev_id": CRED_REV_ID,
            }
        )

        with mock.patch.object(
            test_module.IssuerCredRevRecord,
            "retrieve_by_cred_ex_id",
            mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError("no such rec")
            ),
        ):
            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.get_cred_rev_record(self.request)

    async def test_get_active_rev_reg(self):
        CRED_DEF_ID = f"{self.test_did}:3:CL:1234:default"
        self.request.match_info = {"cred_def_id": CRED_DEF_ID}

        with (
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_indy_revoc.return_value = mock.MagicMock(
                get_active_issuer_rev_reg_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        serialize=mock.MagicMock(return_value="dummy")
                    )
                )
            )

            result = await test_module.get_active_rev_reg(self.request)
            mock_json_response.assert_called_once_with({"result": "dummy"})
            assert result is mock_json_response.return_value

    async def test_get_active_rev_reg_not_found(self):
        CRED_DEF_ID = f"{self.test_did}:3:CL:1234:default"
        self.request.match_info = {"cred_def_id": CRED_DEF_ID}

        with (
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_indy_revoc.return_value = mock.MagicMock(
                get_active_issuer_rev_reg_record=mock.CoroutineMock(
                    side_effect=test_module.StorageNotFoundError(error_code="dummy")
                )
            )

            with self.assertRaises(HTTPNotFound):
                await test_module.get_active_rev_reg(self.request)
            mock_json_response.assert_not_called()

    async def test_get_tails_file(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with (
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "FileResponse", mock.Mock()
            ) as mock_file_response,
        ):
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(tails_local_path="dummy")
                )
            )

            result = await test_module.get_tails_file(self.request)
            mock_file_response.assert_called_once_with(path="dummy", status=200)
            assert result is mock_file_response.return_value

    async def test_get_tails_file_not_found(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with (
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "FileResponse", mock.Mock()
            ) as mock_file_response,
        ):
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    side_effect=test_module.StorageNotFoundError(error_code="dummy")
                )
            )

            with self.assertRaises(HTTPNotFound):
                await test_module.get_tails_file(self.request)
            mock_file_response.assert_not_called()

    async def test_upload_tails_file_basic(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with (
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_upload = mock.CoroutineMock()
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        tails_local_path=f"/tmp/tails/{REV_REG_ID}",
                        has_local_tails_file=True,
                        upload_tails_file=mock_upload,
                    )
                )
            )
            result = await test_module.upload_tails_file(self.request)
            mock_upload.assert_awaited_once()
            mock_json_response.assert_called_once_with({})
            assert result is mock_json_response.return_value

    async def test_upload_tails_file_no_local_tails_file(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with mock.patch.object(
            test_module, "IndyRevocation", autospec=True
        ) as mock_indy_revoc:
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        tails_local_path=f"/tmp/tails/{REV_REG_ID}",
                        has_local_tails_file=False,
                    )
                )
            )

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.upload_tails_file(self.request)

    async def test_upload_tails_file_fail(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with mock.patch.object(
            test_module, "IndyRevocation", autospec=True
        ) as mock_indy_revoc:
            mock_upload = mock.CoroutineMock(side_effect=RevocationError("test"))
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        tails_local_path=f"/tmp/tails/{REV_REG_ID}",
                        has_local_tails_file=True,
                        upload_tails_file=mock_upload,
                    )
                )
            )

            with self.assertRaises(test_module.web.HTTPInternalServerError):
                await test_module.upload_tails_file(self.request)

    async def test_send_rev_reg_def(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with (
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        send_def=mock.CoroutineMock(),
                        send_entry=mock.CoroutineMock(),
                        serialize=mock.MagicMock(return_value="dummy"),
                    )
                )
            )

            result = await test_module.send_rev_reg_def(self.request)
            mock_json_response.assert_called_once_with({"sent": "dummy"})
            assert result is mock_json_response.return_value

    async def test_send_rev_reg_def_not_found(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with (
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "FileResponse", mock.Mock()
            ) as mock_json_response,
        ):
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    side_effect=test_module.StorageNotFoundError(error_code="dummy")
                )
            )

            with self.assertRaises(HTTPNotFound):
                await test_module.send_rev_reg_def(self.request)
            mock_json_response.assert_not_called()

    async def test_send_rev_reg_def_x(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with mock.patch.object(
            test_module, "IndyRevocation", autospec=True
        ) as mock_indy_revoc:
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        send_def=mock.CoroutineMock(
                            side_effect=test_module.RevocationError()
                        ),
                    )
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.send_rev_reg_def(self.request)

    async def test_send_rev_reg_entry(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with (
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        send_entry=mock.CoroutineMock(),
                        serialize=mock.MagicMock(return_value="dummy"),
                    )
                )
            )

            result = await test_module.send_rev_reg_entry(self.request)
            mock_json_response.assert_called_once_with({"result": "dummy"})
            assert result is mock_json_response.return_value

    async def test_send_rev_reg_entry_not_found(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with (
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "FileResponse", mock.Mock()
            ) as mock_json_response,
        ):
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    side_effect=test_module.StorageNotFoundError(error_code="dummy")
                )
            )

            with self.assertRaises(HTTPNotFound):
                await test_module.send_rev_reg_entry(self.request)
            mock_json_response.assert_not_called()

    async def test_send_rev_reg_entry_x(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}

        with mock.patch.object(
            test_module, "IndyRevocation", autospec=True
        ) as mock_indy_revoc:
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        send_entry=mock.CoroutineMock(
                            side_effect=test_module.RevocationError()
                        ),
                    )
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.send_rev_reg_entry(self.request)

    async def test_update_rev_reg(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}
        self.request.json = mock.CoroutineMock(
            return_value={"tails_public_uri": f"http://sample.ca:8181/tails/{REV_REG_ID}"}
        )

        with (
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        set_tails_file_public_uri=mock.CoroutineMock(),
                        save=mock.CoroutineMock(),
                        serialize=mock.MagicMock(return_value="dummy"),
                    )
                )
            )

            result = await test_module.update_rev_reg(self.request)
            mock_json_response.assert_called_once_with({"result": "dummy"})
            assert result is mock_json_response.return_value

    async def test_update_rev_reg_not_found(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}
        self.request.json = mock.CoroutineMock(
            return_value={"tails_public_uri": f"http://sample.ca:8181/tails/{REV_REG_ID}"}
        )

        with (
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "FileResponse", mock.Mock()
            ) as mock_json_response,
        ):
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    side_effect=test_module.StorageNotFoundError(error_code="dummy")
                )
            )

            with self.assertRaises(HTTPNotFound):
                await test_module.update_rev_reg(self.request)
            mock_json_response.assert_not_called()

    async def test_update_rev_reg_x(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}
        self.request.json = mock.CoroutineMock(
            return_value={"tails_public_uri": f"http://sample.ca:8181/tails/{REV_REG_ID}"}
        )

        with mock.patch.object(
            test_module, "IndyRevocation", autospec=True
        ) as mock_indy_revoc:
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        set_tails_file_public_uri=mock.CoroutineMock(
                            side_effect=test_module.RevocationError()
                        ),
                    )
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.update_rev_reg(self.request)

    async def test_set_rev_reg_state(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}
        self.request.json = mock.CoroutineMock(
            return_value={
                "max_cred_num": "1000",
            }
        )
        self.request.query = {
            "state": test_module.IssuerRevRegRecord.STATE_ACTIVE,
        }

        with (
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        set_state=mock.CoroutineMock(),
                        save=mock.CoroutineMock(),
                        serialize=mock.MagicMock(return_value="dummy"),
                    )
                )
            )

            result = await test_module.set_rev_reg_state(self.request)
            mock_json_response.assert_called_once_with({"result": "dummy"})
            assert result is mock_json_response.return_value

    async def test_set_rev_reg_state_not_found(self):
        REV_REG_ID = "{}:4:{}:3:CL:1234:default:CL_ACCUM:default".format(
            self.test_did, self.test_did
        )
        self.request.match_info = {"rev_reg_id": REV_REG_ID}
        self.request.json = mock.CoroutineMock(
            return_value={
                "max_cred_num": "1000",
            }
        )
        self.request.query = {
            "state": test_module.IssuerRevRegRecord.STATE_ACTIVE,
        }

        with (
            mock.patch.object(
                test_module, "IndyRevocation", autospec=True
            ) as mock_indy_revoc,
            mock.patch.object(
                test_module.web, "FileResponse", mock.Mock()
            ) as mock_json_response,
        ):
            mock_indy_revoc.return_value = mock.MagicMock(
                get_issuer_rev_reg_record=mock.CoroutineMock(
                    side_effect=test_module.StorageNotFoundError(error_code="dummy")
                )
            )

            with self.assertRaises(HTTPNotFound):
                await test_module.set_rev_reg_state(self.request)
            mock_json_response.assert_not_called()

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(
            _state={
                "swagger_dict": {
                    "paths": {
                        "/revocation/registry/{rev_reg_id}/tails-file": {
                            "get": {"responses": {"200": {"description": "tails file"}}}
                        }
                    }
                }
            }
        )
        test_module.post_process_routes(mock_app)
        assert mock_app._state["swagger_dict"]["paths"][
            "/revocation/registry/{rev_reg_id}/tails-file"
        ]["get"]["responses"]["200"]["schema"] == {"type": "string", "format": "binary"}

        assert "tags" in mock_app._state["swagger_dict"]

    @mock.patch.object(
        IssuerRevRegRecord, "retrieve_by_revoc_reg_id", return_value=mock.MagicMock()
    )
    @mock.patch.object(
        RevocationManager, "update_rev_reg_revoked_state", return_value=(True, True, True)
    )
    async def test_update_rev_reg_revoked_state(self, *_):
        self.request.query = {"apply_ledger_update": "true"}
        self.request.match_info = {"rev_reg_id": "rev_reg_id"}

        mock_ledger_manager = mock.MagicMock(BaseMultipleLedgerManager, autospec=True)
        mock_ledger_manager.get_write_ledgers = mock.CoroutineMock("ledger")
        self.context.injector.bind_instance(
            BaseMultipleLedgerManager, mock_ledger_manager
        )

        mock_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_ledger.pool = mock.MagicMock(genesis_txns="genesis_txns")
        self.context.injector.bind_instance(BaseLedger, mock_ledger)
        self.profile.context.injector.bind_instance(BaseLedger, mock_ledger)

        result = await test_module.update_rev_reg_revoked_state(self.request)
        assert result.status == 200


def make_mock_request(query=None, path="/admin/revocation/delete_tails"):
    query = query or {}
    request = mock.MagicMock()
    request.__getitem__.side_effect = lambda key: {
        "context": mock.MagicMock(profile=mock.MagicMock())
    }[key]
    request.headers = {"Authorization": "Bearer fake-token"}
    request.query = MultiDict(query)
    request.path = path
    return request


@pytest.mark.asyncio
class TestDeleteTails:
    def setup_method(self):
        self.rev_reg_id = "rev_reg_id_123"
        self.cred_def_id = "cred_def_id_456"
        self.main_dir_rev = "path/to/main/dir/rev"
        os.makedirs(self.main_dir_rev, exist_ok=True)
        tails_file = os.path.join(self.main_dir_rev, "tails")
        with open(tails_file, "w") as f:
            f.write("test tails file")

    @mock.patch("acapy_agent.revocation.routes.IndyRevocation.get_issuer_rev_reg_record")
    @mock.patch("acapy_agent.revocation.routes.shutil.rmtree")
    async def test_delete_tails_by_rev_reg_id(self, mock_rmtree, mock_get_rev_reg_record):
        tails_file_path = os.path.join(self.main_dir_rev, "tails")
        mock_record = mock.AsyncMock()
        mock_record.tails_local_path = tails_file_path

        mock_get_rev_reg_record.return_value = mock_record

        request = make_mock_request({"rev_reg_id": self.rev_reg_id})

        result = await test_module.delete_tails(request)

        mock_rmtree.assert_called_once_with(self.main_dir_rev)

        body_bytes = result.body
        body = json.loads(body_bytes.decode("utf-8"))

        assert "message" in body
        assert body["message"] == "All files deleted successfully"

    @mock.patch("acapy_agent.revocation.routes.os.listdir")
    @mock.patch("acapy_agent.revocation.routes.IssuerRevRegRecord.query_by_cred_def_id")
    @mock.patch("acapy_agent.revocation.routes.shutil.rmtree")
    async def test_delete_tails_by_cred_def_id(
        self, mock_rmtree, mock_query_by_cred_def_id, mock_listdir
    ):
        main_dir_cred = "/path/to/main/dir"
        cred_def_id = self.cred_def_id
        cred_dir_name = f"{cred_def_id}_folder"

        mock_listdir.return_value = [cred_dir_name, "other_folder"]

        record = mock.Mock()
        record.tails_local_path = os.path.join(main_dir_cred, cred_dir_name, "tails")
        mock_query_by_cred_def_id.return_value = [record]

        request = make_mock_request({"cred_def_id": cred_def_id})

        result = await test_module.delete_tails(request)

        expected_rmtree_path = os.path.join(main_dir_cred, cred_dir_name)
        mock_rmtree.assert_called_once_with(expected_rmtree_path)

        body_bytes = result.body
        body = json.loads(body_bytes.decode("utf-8"))

        assert "message" in body
        assert body["message"] == "All files deleted successfully"

    @mock.patch("acapy_agent.revocation.routes.IssuerRevRegRecord.query_by_cred_def_id")
    @mock.patch("acapy_agent.revocation.routes.os.listdir")
    async def test_delete_tails_not_found(self, mock_listdir, mock_query_by_cred_def_id):
        mock_query_by_cred_def_id.return_value = []
        mock_listdir.return_value = []  # Important! Avoid list index error.

        request = make_mock_request({"cred_def_id": "nonexistent_cred_def_id"})

        result = await test_module.delete_tails(request)

        if hasattr(result, "json"):
            body = await result.json()
        elif hasattr(result, "body"):
            body = json.loads(result.body.decode())
        else:
            body = result

        assert "message" in body

    def teardown_method(self):
        if os.path.exists(self.main_dir_rev):
            shutil.rmtree(self.main_dir_rev)



"""V2.0 issue-credential linked data proof credential format handler."""

import logging
from typing import Mapping, Optional

from marshmallow import EXCLUDE, INCLUDE
from pyld import jsonld
from pyld.jsonld import JsonLdProcessor

from acapy_agent.wallet.default_verification_key_strategy import (
    BaseVerificationKeyStrategy,
    VerificationKeyStrategyError,
)

from ......messaging.decorators.attach_decorator import AttachDecorator
from ......storage.vc_holder.base import VCHolder
from ......storage.vc_holder.vc_record import VCRecord
from ......vc.ld_proofs import DocumentLoader
from ......vc.ld_proofs.check import get_properties_without_context
from ......vc.ld_proofs.error import LinkedDataProofException
from ......vc.vc_ld import VerifiableCredential, VerifiableCredentialSchema
from ......vc.vc_ld.manager import VcLdpManager, VcLdpManagerError
from ......vc.vc_ld.models.options import LDProofVCOptions
from ...message_types import (
    ATTACHMENT_FORMAT,
    CRED_20_ISSUE,
    CRED_20_OFFER,
    CRED_20_PROPOSAL,
    CRED_20_REQUEST,
)
from ...messages.cred_format import V20CredFormat
from ...messages.cred_issue import V20CredIssue
from ...messages.cred_offer import V20CredOffer
from ...messages.cred_proposal import V20CredProposal
from ...messages.cred_request import V20CredRequest
from ...models.cred_ex_record import V20CredExRecord
from ...models.detail.ld_proof import V20CredExRecordLDProof
from ..handler import CredFormatAttachment, V20CredFormatError, V20CredFormatHandler
from .models.cred_detail import LDProofVCDetail, LDProofVCDetailSchema

LOGGER = logging.getLogger(__name__)


class LDProofCredFormatHandler(V20CredFormatHandler):
    """Linked data proof credential format handler."""

    format = V20CredFormat.Format.LD_PROOF

    @classmethod
    def validate_fields(cls, message_type: str, attachment_data: Mapping) -> None:
        """Validate attachment data for a specific message type.

        Uses marshmallow schemas to validate if format specific attachment data
        is valid for the specified message type. Only does structural and type
        checks, does not validate if .e.g. the issuer value is valid.


        Args:
            message_type (str): The message type to validate the attachment data for.
                Should be one of the message types as defined in message_types.py
            attachment_data (Mapping): [description]
                The attachment data to validate

        Raises:
            Exception: When the data is not valid.

        """
        mapping = {
            CRED_20_PROPOSAL: LDProofVCDetailSchema,
            CRED_20_OFFER: LDProofVCDetailSchema,
            CRED_20_REQUEST: LDProofVCDetailSchema,
            CRED_20_ISSUE: VerifiableCredentialSchema,
        }

        # Get schema class
        Schema = mapping[message_type]

        # Validate, throw if not valid
        Schema(unknown=EXCLUDE).load(attachment_data)

    async def get_detail_record(self, cred_ex_id: str) -> V20CredExRecordLDProof:
        """Retrieve credential exchange detail record by cred_ex_id."""

        async with self.profile.session() as session:
            records = await LDProofCredFormatHandler.format.detail.query_by_cred_ex_id(
                session, cred_ex_id
            )

        if len(records) > 1:
            LOGGER.warning(
                "Cred ex id %s has %d %s detail records: should be 1",
                cred_ex_id,
                len(records),
                LDProofCredFormatHandler.format.api,
            )
        return records[0] if records else None

    def get_format_identifier(self, message_type: str) -> str:
        """Get attachment format identifier for format and message combination.

        Args:
            message_type (str): Message type for which to return the format identifier

        Returns:
            str: Issue credential attachment format identifier

        """
        return ATTACHMENT_FORMAT[message_type][LDProofCredFormatHandler.format.api]

    def get_format_data(self, message_type: str, data: dict) -> CredFormatAttachment:
        """Get credential format and attachment objects for use in cred ex messages.

        Returns a tuple of both credential format and attachment decorator for use
        in credential exchange messages. It looks up the correct format identifier and
        encodes the data as a base64 attachment.

        Args:
            message_type (str): The message type for which to return the cred format.
                Should be one of the message types defined in the message types file
            data (dict): The data to include in the attach decorator

        Returns:
            CredFormatAttachment: Credential format and attachment data objects

        """
        return (
            V20CredFormat(
                attach_id=LDProofCredFormatHandler.format.api,
                format_=self.get_format_identifier(message_type),
            ),
            AttachDecorator.data_base64(data, ident=LDProofCredFormatHandler.format.api),
        )

    async def create_proposal(
        self, cred_ex_record: V20CredExRecord, proposal_data: Mapping
    ) -> CredFormatAttachment:
        """Create linked data proof credential proposal."""
        manager = VcLdpManager(self.profile)
        detail = LDProofVCDetail.deserialize(proposal_data)
        assert detail.options and isinstance(detail.options, LDProofVCOptions)
        assert detail.credential and isinstance(detail.credential, VerifiableCredential)
        try:
            detail.credential = await manager.prepare_credential(
                detail.credential, detail.options
            )
        except VcLdpManagerError as err:
            raise V20CredFormatError("Failed to prepare credential") from err

        return self.get_format_data(CRED_20_PROPOSAL, detail.serialize())

    async def receive_proposal(
        self, cred_ex_record: V20CredExRecord, cred_proposal_message: V20CredProposal
    ) -> None:
        """Receive linked data proof credential proposal."""

    async def create_offer(
        self, cred_proposal_message: V20CredProposal
    ) -> CredFormatAttachment:
        """Create linked data proof credential offer."""
        if not cred_proposal_message:
            raise V20CredFormatError(
                "Cannot create linked data proof offer without proposal data"
            )

        # Parse offer data which is either a proposal or an offer.
        # Data is stored in proposal if we received a proposal
        # but also when we create an offer (manager does some weird stuff)
        offer_data = cred_proposal_message.attachment(LDProofCredFormatHandler.format)
        detail = LDProofVCDetail.deserialize(offer_data)
        manager = VcLdpManager(self.profile)
        assert detail.options and isinstance(detail.options, LDProofVCOptions)
        assert detail.credential and isinstance(detail.credential, VerifiableCredential)
        try:
            detail.credential = await manager.prepare_credential(
                detail.credential, detail.options
            )
        except VcLdpManagerError as err:
            raise V20CredFormatError("Failed to prepare credential") from err

        document_loader = self.profile.inject(DocumentLoader)
        missing_properties = get_properties_without_context(
            detail.credential.serialize(), document_loader
        )

        if len(missing_properties) > 0:
            raise LinkedDataProofException(
                f"{len(missing_properties)} attributes dropped. "
                f"Provide definitions in context to correct. {missing_properties}"
            )

        # Make sure we can issue with the did and proof type
        try:
            # Check suitable verification method for signing (fails if none suitable)
            verkey_id_strategy = self.profile.context.inject(BaseVerificationKeyStrategy)
            _ = await verkey_id_strategy.get_verification_method_id_for_did(
                detail.credential.issuer_id,
                self.profile,
                proof_type=detail.options.proof_type,
                proof_purpose="assertionMethod",
                verification_method_id=detail.options.verification_method,
            )
        except VerificationKeyStrategyError as err:
            raise V20CredFormatError(
                "Checking whether issuance is possible failed"
            ) from err

        return self.get_format_data(CRED_20_OFFER, detail.serialize())

    async def receive_offer(
        self, cred_ex_record: V20CredExRecord, cred_offer_message: V20CredOffer
    ) -> None:
        """Receive linked data proof credential offer."""

    async def create_request(
        self, cred_ex_record: V20CredExRecord, request_data: Optional[Mapping] = None
    ) -> CredFormatAttachment:
        """Create linked data proof credential request."""
        holder_did = request_data.get("holder_did") if request_data else None

        if cred_ex_record.cred_offer:
            request_data = cred_ex_record.cred_offer.attachment(
                LDProofCredFormatHandler.format
            )
        # API data is stored in proposal (when starting from request)
        # It is a bit of a strange flow IMO.
        elif cred_ex_record.cred_proposal:
            request_data = cred_ex_record.cred_proposal.attachment(
                LDProofCredFormatHandler.format
            )
        else:
            raise V20CredFormatError(
                "Cannot create linked data proof request without offer or input data"
            )

        detail = LDProofVCDetail.deserialize(request_data)
        manager = VcLdpManager(self.profile)
        assert detail.options and isinstance(detail.options, LDProofVCOptions)
        assert detail.credential and isinstance(detail.credential, VerifiableCredential)
        try:
            detail.credential = await manager.prepare_credential(
                detail.credential, detail.options, holder_did=holder_did
            )
        except VcLdpManagerError as err:
            raise V20CredFormatError("Failed to prepare credential") from err

        return self.get_format_data(CRED_20_REQUEST, detail.serialize())

    def can_receive_request_without_offer(
        self,
    ) -> bool:
        """Can this handler receive credential request without an offer?"""
        return True

    async def receive_request(
        self, cred_ex_record: V20CredExRecord, cred_request_message: V20CredRequest
    ) -> None:
        """Receive linked data proof request."""
        # Check that request hasn't substantially changed from offer, if offer sent
        if cred_ex_record.cred_offer:
            offer_detail_dict = cred_ex_record.cred_offer.attachment(
                LDProofCredFormatHandler.format
            )
            req_detail_dict = cred_request_message.attachment(
                LDProofCredFormatHandler.format
            )

            # If credentialSubject.id in offer, it should be the same in request
            offer_id = (
                offer_detail_dict["credential"].get("credentialSubject", {}).get("id")
            )
            request_id = (
                req_detail_dict["credential"].get("credentialSubject", {}).get("id")
            )
            if offer_id and offer_id != request_id:
                raise V20CredFormatError(
                    "Request credentialSubject.id must match offer credentialSubject.id"
                )

            # Nothing else should be different about the request
            if request_id:
                offer_detail_dict["credential"].setdefault("credentialSubject", {})[
                    "id"
                ] = request_id
            if offer_detail_dict != req_detail_dict:
                raise V20CredFormatError("Request must match offer if offer is sent")

    async def issue_credential(
        self,
        cred_ex_record: V20CredExRecord,
        retries: int = 5,
    ) -> CredFormatAttachment:
        """Issue linked data proof credential."""
        if not cred_ex_record.cred_request:
            raise V20CredFormatError("Cannot issue credential without credential request")

        detail_dict = cred_ex_record.cred_request.attachment(
            LDProofCredFormatHandler.format
        )
        detail = LDProofVCDetail.deserialize(detail_dict)
        manager = VcLdpManager(self.profile)
        assert detail.options and isinstance(detail.options, LDProofVCOptions)
        assert detail.credential and isinstance(detail.credential, VerifiableCredential)
        try:
            vc = await manager.issue(detail.credential, detail.options)
        except VcLdpManagerError as err:
            raise V20CredFormatError("Failed to issue credential") from err

        return self.get_format_data(CRED_20_ISSUE, vc.serialize())

    async def receive_credential(
        self, cred_ex_record: V20CredExRecord, cred_issue_message: V20CredIssue
    ) -> None:
        """Receive linked data proof credential."""
        cred_dict = cred_issue_message.attachment(LDProofCredFormatHandler.format)
        detail_dict = cred_ex_record.cred_request.attachment(
            LDProofCredFormatHandler.format
        )

        vc = VerifiableCredential.deserialize(cred_dict, unknown=INCLUDE)
        detail = LDProofVCDetail.deserialize(detail_dict)

        # Remove values from cred that are not part of detail
        cred_dict.pop("proof")
        credential_status = cred_dict.get("credentialStatus", None)
        detail_status = detail.options.credential_status

        if cred_dict != detail_dict["credential"]:
            raise V20CredFormatError(
                f"Received credential for cred_ex_id {cred_ex_record.cred_ex_id} does not"
                " match requested credential"
            )

        # both credential and detail contain status. Check for equality
        if credential_status and detail_status:
            if credential_status.get("type") != detail_status.get("type"):
                raise V20CredFormatError(
                    "Received credential status type does not match credential request"
                )
        # Either credential or detail contains status. Throw error
        elif (credential_status and not detail_status) or (
            not credential_status and detail_status
        ):
            raise V20CredFormatError(
                "Received credential status contains credential status"
                " that does not match credential request"
            )

        # TODO: if created wasn't present in the detail options, should we verify
        # it is ~now (e.g. some time in the past + future)?
        # Check if created property matches
        if detail.options.created and vc.proof.created != detail.options.created:
            raise V20CredFormatError(
                "Received credential proof.created does not"
                " match options.created from credential request"
            )

        # Check challenge
        if vc.proof.challenge != detail.options.challenge:
            raise V20CredFormatError(
                "Received credential proof.challenge does not"
                " match options.challenge from credential request"
            )

        # Check domain
        if vc.proof.domain != detail.options.domain:
            raise V20CredFormatError(
                "Received credential proof.domain does not"
                " match options.domain from credential request"
            )

        # Check if proof type matches
        if vc.proof.type != detail.options.proof_type:
            raise V20CredFormatError(
                "Received credential proof.type does not"
                " match options.proofType from credential request"
            )

    async def store_credential(
        self, cred_ex_record: V20CredExRecord, cred_id: Optional[str] = None
    ) -> None:
        """Store linked data proof credential."""
        # Get attachment data
        cred_dict: dict = cred_ex_record.cred_issue.attachment(
            LDProofCredFormatHandler.format
        )

        # Deserialize objects
        credential = VerifiableCredential.deserialize(cred_dict, unknown=INCLUDE)

        # Get signature suite, proof purpose and document loader
        manager = VcLdpManager(self.profile)
        try:
            result = await manager.verify_credential(credential)
        except VcLdpManagerError as err:
            raise V20CredFormatError("Failed to verify credential") from err

        if not result.verified:
            raise V20CredFormatError(f"Received invalid credential: {result}")

        # Saving expanded type as a cred_tag
        document_loader = self.profile.inject(DocumentLoader)
        expanded = jsonld.expand(cred_dict, options={"documentLoader": document_loader})
        types = JsonLdProcessor.get_values(
            expanded[0],
            "@type",
        )

        # create VC record for storage
        vc_record = VCRecord(
            contexts=credential.context_urls,
            expanded_types=types,
            issuer_id=credential.issuer_id,
            subject_ids=credential.credential_subject_ids,
            schema_ids=[],  # Schemas not supported yet
            proof_types=[credential.proof.type],
            cred_value=credential.serialize(),
            given_id=credential.id,
            record_id=cred_id,
            cred_tags=None,  # Tags should be derived from credential values
        )

        # Create detail record with cred_id_stored
        detail_record = V20CredExRecordLDProof(
            cred_ex_id=cred_ex_record.cred_ex_id, cred_id_stored=vc_record.record_id
        )

        # save credential and detail record
        async with self.profile.session() as session:
            vc_holder = session.inject(VCHolder)

            await vc_holder.store_credential(vc_record)
            # Store detail record, emit event
            await detail_record.save(session, reason="store credential v2.0", event=True)



"""DIDComm prefix management."""

import re
from enum import Enum

QUALIFIED = re.compile(r"^[a-zA-Z\-\+]+:.+")


def qualify(msg_type: str, prefix: str):
    """Qualify a message type with a prefix, if unqualified."""

    return msg_type if QUALIFIED.match(msg_type or "") else f"{prefix}/{msg_type}"


class DIDCommPrefix(Enum):
    """Enum for DIDComm Prefix, old or new style, per Aries RFC 384."""

    NEW = "https://didcomm.org"
    OLD = "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec"

    def qualify(self, msg_type: str) -> str:
        """Qualify input message type with prefix and separator."""

        return qualify(msg_type, self.value)

    @classmethod
    def qualify_all(cls, messages: dict) -> dict:
        """Apply all known prefixes to a dictionary of message types."""

        return {qualify(k, pfx.value): v for pfx in cls for k, v in messages.items()}

    @staticmethod
    def qualify_current(slug: str) -> str:
        """Qualify input slug with prefix currently in effect and separator.

        This method now will always use the new prefix.
        """

        return qualify(slug, DIDCommPrefix.NEW.value)

    @staticmethod
    def unqualify(qual: str) -> str:
        """Strip prefix and separator from input, if present, and return result."""
        for pfx in DIDCommPrefix:
            if (qual or "").startswith(f"{pfx.value}/"):
                return qual.split(f"{pfx.value}/")[1]

        return qual



"""Manager for performing Linked Data Proof signatures over JSON-LD formatted W3C VCs."""

from datetime import datetime, timezone
from typing import Dict, List, Optional, Type, Union, cast

from acapy_agent.wallet.keys.manager import MultikeyManager, multikey_to_verkey
from pyld import jsonld
from pyld.jsonld import JsonLdProcessor

from acapy_agent.vc.ld_proofs.suites.ecdsa_secp256r1_signature_2019 import (
    EcdsaSecp256r1Signature2019,
)

from ...core.profile import Profile
from ...storage.vc_holder.base import VCHolder
from ...storage.vc_holder.vc_record import VCRecord
from ...wallet.base import BaseWallet
from ...wallet.default_verification_key_strategy import BaseVerificationKeyStrategy
from ...wallet.did_info import DIDInfo
from ...wallet.key_type import BLS12381G2, ED25519, P256, KeyType
from ..ld_proofs.constants import (
    CREDENTIALS_CONTEXT_V1_URL,
    CREDENTIALS_CONTEXT_V2_URL,
    SECURITY_CONTEXT_BBS_URL,
    SECURITY_CONTEXT_ED25519_2020_URL,
)
from ..ld_proofs.crypto.wallet_key_pair import WalletKeyPair
from ..ld_proofs.document_loader import DocumentLoader
from ..ld_proofs.purposes.authentication_proof_purpose import AuthenticationProofPurpose
from ..ld_proofs.purposes.credential_issuance_purpose import CredentialIssuancePurpose
from ..ld_proofs.purposes.proof_purpose import ProofPurpose
from ..ld_proofs.suites.bbs_bls_signature_2020 import BbsBlsSignature2020
from ..ld_proofs.suites.bbs_bls_signature_proof_2020 import BbsBlsSignatureProof2020
from ..ld_proofs.suites.ed25519_signature_2018 import Ed25519Signature2018
from ..ld_proofs.suites.ed25519_signature_2020 import Ed25519Signature2020
from ..ld_proofs.suites.linked_data_proof import LinkedDataProof
from ..ld_proofs.validation_result import DocumentVerificationResult
from ..vc_ld.models.presentation import VerifiablePresentation
from ..vc_ld.validation_result import PresentationVerificationResult
from .external_suite import ExternalSuiteNotFoundError, ExternalSuiteProvider
from .issue import issue as ldp_issue
from .models.credential import VerifiableCredential
from .models.linked_data_proof import LDProof
from .models.options import LDProofVCOptions
from .prove import sign_presentation
from .verify import verify_credential, verify_presentation

SignatureTypes = Union[
    Type[Ed25519Signature2018],
    Type[Ed25519Signature2020],
    Type[EcdsaSecp256r1Signature2019],
    Type[BbsBlsSignature2020],
]
ProofTypes = Union[
    Type[Ed25519Signature2018],
    Type[Ed25519Signature2020],
    Type[EcdsaSecp256r1Signature2019],
    Type[BbsBlsSignature2020],
    Type[BbsBlsSignatureProof2020],
]
SUPPORTED_ISSUANCE_PROOF_PURPOSES = {
    CredentialIssuancePurpose.term,
    AuthenticationProofPurpose.term,
}
SUPPORTED_V2_ISSUANCE_PROOF_TYPES = [
    Ed25519Signature2020.signature_type,
    BbsBlsSignature2020.signature_type,
]
SIGNATURE_SUITE_KEY_TYPE_MAPPING: Dict[SignatureTypes, KeyType] = {
    Ed25519Signature2018: ED25519,
    Ed25519Signature2020: ED25519,
    EcdsaSecp256r1Signature2019: P256,
}
PROOF_KEY_TYPE_MAPPING = cast(Dict[ProofTypes, KeyType], SIGNATURE_SUITE_KEY_TYPE_MAPPING)


# We only want to add bbs suites to supported if the module is installed
if BbsBlsSignature2020.BBS_SUPPORTED:
    SIGNATURE_SUITE_KEY_TYPE_MAPPING[BbsBlsSignature2020] = BLS12381G2
    PROOF_KEY_TYPE_MAPPING[BbsBlsSignatureProof2020] = BLS12381G2


PROOF_TYPE_SIGNATURE_SUITE_MAPPING: Dict[str, SignatureTypes] = {
    suite.signature_type: suite for suite in SIGNATURE_SUITE_KEY_TYPE_MAPPING
}


# key_type -> set of signature types mappings
KEY_TYPE_SIGNATURE_TYPE_MAPPING = {
    key_type: {
        suite.signature_type
        for suite, kt in SIGNATURE_SUITE_KEY_TYPE_MAPPING.items()
        if kt == key_type
    }
    for key_type in SIGNATURE_SUITE_KEY_TYPE_MAPPING.values()
}


class VcLdpManagerError(Exception):
    """Generic VcLdpManager Error."""


class VcLdpManager:
    """Class for managing Linked Data Proof signatures over JSON-LD formatted W3C VCs."""

    def __init__(self, profile: Profile):
        """Initialize the VC LD Proof Manager."""
        self.profile = profile

    async def _did_info_for_did(self, did: str) -> DIDInfo:
        """Get the did info for specified did.

        If the did starts with did:sov it will remove the prefix for
        backwards compatibility with not fully qualified did.

        Args:
            did (str): The did to retrieve from the wallet.

        Raises:
            WalletNotFoundError: If the did is not found in the wallet.

        Returns:
            DIDInfo: did information

        """
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)

            # If the did starts with did:sov we need to query without
            if did.startswith("did:sov:"):
                return await wallet.get_local_did(did.replace("did:sov:", ""))

            # All other methods we can just query
            return await wallet.get_local_did(did)

    async def _get_suite(
        self,
        *,
        proof_type: str,
        verification_method: str,
        proof: dict,
        did_info: DIDInfo,
    ):
        """Get signature suite for issuance of verification."""
        # Try to get suite from external provider first
        try:
            if (provider := self.profile.inject_or(ExternalSuiteProvider)) and (
                suite := await provider.get_suite(
                    self.profile, proof_type, proof, verification_method, did_info
                )
            ):
                return suite
        except ExternalSuiteNotFoundError as error:
            raise VcLdpManagerError(
                f"Unable to get signature suite for proof type {proof_type} "
                "using external provider."
            ) from error

        async with self.profile.session() as session:
            key_manager = MultikeyManager(session)
            key_info = await key_manager.resolve_and_bind_kid(verification_method)
            multikey = key_info["multikey"]

        # Get signature class based on proof type
        SignatureClass = PROOF_TYPE_SIGNATURE_SUITE_MAPPING[proof_type]

        # Generically create signature class
        return SignatureClass(
            verification_method=verification_method,
            proof=proof,
            key_pair=WalletKeyPair(
                profile=self.profile,
                key_type=SIGNATURE_SUITE_KEY_TYPE_MAPPING[SignatureClass],
                public_key_base58=multikey_to_verkey(multikey),
            ),
        )

    def _get_proof_purpose(
        self,
        *,
        proof_purpose: Optional[str] = None,
        challenge: Optional[str] = None,
        domain: Optional[str] = None,
    ) -> ProofPurpose:
        """Get the proof purpose for a credential.

        Args:
            proof_purpose (str): The proof purpose string value
            challenge (str, optional): Challenge
            domain (str, optional): domain

        Raises:
            VcLdpManagerError:
                - If the proof purpose is not supported.
                - [authentication] If challenge is missing.

        Returns:
            ProofPurpose: Proof purpose instance that can be used for issuance.

        """
        # Default proof purpose is assertionMethod
        proof_purpose = proof_purpose or CredentialIssuancePurpose.term

        if proof_purpose == CredentialIssuancePurpose.term:
            return CredentialIssuancePurpose()
        elif proof_purpose == AuthenticationProofPurpose.term:
            # assert challenge is present for authentication proof purpose
            if not challenge:
                raise VcLdpManagerError(
                    f"Challenge is required for '{proof_purpose}' proof purpose."
                )

            return AuthenticationProofPurpose(challenge=challenge, domain=domain)
        else:
            raise VcLdpManagerError(
                f"Unsupported proof purpose: {proof_purpose}. "
                f"Supported proof types are: {SUPPORTED_ISSUANCE_PROOF_PURPOSES}"
            )

    async def prepare_credential(
        self,
        credential: VerifiableCredential,
        options: LDProofVCOptions,
        holder_did: Optional[str] = None,
    ) -> VerifiableCredential:
        """Prepare a credential for issuance."""
        # Limit VCDM 2.0 with Ed25519Signature2020
        if (
            credential.context_urls[0] == CREDENTIALS_CONTEXT_V2_URL
            and options.proof_type not in SUPPORTED_V2_ISSUANCE_PROOF_TYPES
        ):
            raise VcLdpManagerError(
                f"Supported VC 2.0 proof types are: {SUPPORTED_V2_ISSUANCE_PROOF_TYPES}."
            )

        # Add BBS context if not present yet
        if (
            options.proof_type == BbsBlsSignature2020.signature_type
            and SECURITY_CONTEXT_BBS_URL not in credential.context_urls
        ):
            credential.add_context(SECURITY_CONTEXT_BBS_URL)
        # Add ED25519-2020 context if not present yet
        elif (
            options.proof_type == Ed25519Signature2020.signature_type
            and SECURITY_CONTEXT_ED25519_2020_URL not in credential.context_urls
        ):
            credential.add_context(SECURITY_CONTEXT_ED25519_2020_URL)

        # Permit late binding of credential subject:
        # IFF credential subject doesn't already have an id, add holder_did as
        # credentialSubject.id (if provided)
        subject = credential.credential_subject

        # TODO if credential subject is a list, we're only binding the first...
        # How should this be handled?
        if isinstance(subject, list):
            subject = subject[0]

        if (
            not credential.issuance_date
            and credential.context_urls[0] == CREDENTIALS_CONTEXT_V1_URL
        ):
            credential.issuance_date = str(
                datetime.now(timezone.utc).isoformat("T", "seconds")
            )

        if not subject:
            raise VcLdpManagerError("Credential subject is required")

        if holder_did and holder_did.startswith("did:key") and "id" not in subject:
            subject["id"] = holder_did

        return credential

    async def prepare_presentation(
        self,
        presentation: VerifiablePresentation,
        options: LDProofVCOptions,
    ) -> VerifiableCredential:
        """Prepare a presentation for issuance."""
        # Add BBS context if not present yet
        if (
            options.proof_type == BbsBlsSignature2020.signature_type
            and SECURITY_CONTEXT_BBS_URL not in presentation.context_urls
        ):
            presentation.add_context(SECURITY_CONTEXT_BBS_URL)
        # Add ED25519-2020 context if not present yet
        elif (
            options.proof_type == Ed25519Signature2020.signature_type
            and SECURITY_CONTEXT_ED25519_2020_URL not in presentation.context_urls
        ):
            presentation.add_context(SECURITY_CONTEXT_ED25519_2020_URL)

        return presentation

    async def _get_suite_for_document(
        self,
        document: Union[VerifiableCredential, VerifiablePresentation],
        options: LDProofVCOptions,
    ) -> LinkedDataProof:
        if isinstance(document, VerifiableCredential):
            issuer_id = document.issuer_id
        elif isinstance(document, VerifiablePresentation):
            issuer_id = document.holder_id

        proof_type = options.proof_type

        if not issuer_id:
            raise VcLdpManagerError("Credential issuer id is required")

        if not proof_type:
            raise VcLdpManagerError("Proof type is required")

        # Create base proof object with options
        proof = LDProof(
            created=options.created,
            domain=options.domain,
            challenge=options.challenge,
        )

        did_info = await self._did_info_for_did(issuer_id)

        # Determine/check suitable verification method for signing. fails if none suitable
        verkey_id_strategy = self.profile.context.inject(BaseVerificationKeyStrategy)
        verification_method_id = (
            await verkey_id_strategy.get_verification_method_id_for_did(
                issuer_id,
                self.profile,
                proof_type=proof_type,
                proof_purpose="assertionMethod",
                verification_method_id=options.verification_method,
            )
        )

        suite = await self._get_suite(
            proof_type=proof_type,
            verification_method=verification_method_id,
            proof=proof.serialize(),
            did_info=did_info,
        )

        return suite

    async def _get_all_proof_suites(self) -> List[LinkedDataProof]:
        """Get all supported suites for verifying presentation.

        Returns a list of suites instantiated with a key type, derived from
        PROOF_KEY_TYPE_MAPPING entries.
        """
        return [
            # Satisfy type checks with a cast to LinkedDataProof
            cast(
                LinkedDataProof,
                # Instantiate suite with a key type
                SuiteClass(
                    key_pair=WalletKeyPair(profile=self.profile, key_type=key_type),
                ),
            )
            # for each suite class -> key_type pair from PROOF_KEY_TYPE_MAPPING
            for SuiteClass, key_type in PROOF_KEY_TYPE_MAPPING.items()
        ]

    async def issue(
        self, credential: VerifiableCredential, options: LDProofVCOptions
    ) -> VerifiableCredential:
        """Sign a VC with a Linked Data Proof."""
        credential = await self.prepare_credential(credential, options)

        # Get signature suite, proof purpose and document loader
        suite = await self._get_suite_for_document(credential, options)
        proof_purpose = self._get_proof_purpose(
            proof_purpose=options.proof_purpose,
            challenge=options.challenge,
            domain=options.domain,
        )
        document_loader = self.profile.inject(DocumentLoader)

        vc = await ldp_issue(
            credential=credential.serialize(),
            suite=suite,
            document_loader=document_loader,
            purpose=proof_purpose,
        )
        return VerifiableCredential.deserialize(vc)

    async def store_credential(
        self,
        vc: VerifiableCredential,
        cred_id: Optional[str] = None,
    ) -> VCRecord:
        """Store a verifiable credential."""

        # Saving expanded type as a cred_tag
        document_loader = self.profile.inject(DocumentLoader)
        expanded = jsonld.expand(
            vc.serialize(), options={"documentLoader": document_loader}
        )
        types = JsonLdProcessor.get_values(
            expanded[0],
            "@type",
        )
        vc_record = VCRecord(
            contexts=vc.context_urls,
            expanded_types=types,
            issuer_id=vc.issuer_id,
            subject_ids=vc.credential_subject_ids,
            schema_ids=[],  # Schemas not supported yet
            proof_types=[vc.proof.type],
            cred_value=vc.serialize(),
            given_id=vc.id,
            record_id=cred_id,
            cred_tags=None,  # Tags should be derived from credential values
        )

        async with self.profile.session() as session:
            vc_holder = session.inject(VCHolder)

            await vc_holder.store_credential(vc_record)

        return vc_record

    async def verify_credential(
        self, vc: VerifiableCredential
    ) -> DocumentVerificationResult:
        """Verify a VC with a Linked Data Proof."""
        return await verify_credential(
            credential=vc.serialize(),
            suites=await self._get_all_proof_suites(),
            document_loader=self.profile.inject(DocumentLoader),
        )

    async def prove(
        self, presentation: VerifiablePresentation, options: LDProofVCOptions
    ) -> VerifiablePresentation:
        """Sign a VP with a Linked Data Proof."""
        presentation = await self.prepare_presentation(presentation, options)

        # Get signature suite, proof purpose and document loader
        suite = await self._get_suite_for_document(presentation, options)
        proof_purpose = self._get_proof_purpose(
            proof_purpose=options.proof_purpose,
            challenge=options.challenge,
            domain=options.domain,
        )
        document_loader = self.profile.inject(DocumentLoader)

        vp = await sign_presentation(
            presentation=presentation.serialize(),
            suite=suite,
            document_loader=document_loader,
            purpose=proof_purpose,
        )
        return VerifiablePresentation.deserialize(vp)

    async def verify_presentation(
        self, vp: VerifiablePresentation, options: LDProofVCOptions
    ) -> PresentationVerificationResult:
        """Verify a VP with a Linked Data Proof."""

        if not options.challenge:
            raise VcLdpManagerError("Challenge is required for verifying a VP")

        return await verify_presentation(
            presentation=vp.serialize(),
            suites=await self._get_all_proof_suites(),
            document_loader=self.profile.inject(DocumentLoader),
            challenge=options.challenge,
        )



"""Verifiable Credential and Presentation proving methods."""

from typing import List, Optional

from ..ld_proofs import (
    AuthenticationProofPurpose,
    DocumentLoaderMethod,
    LinkedDataProof,
    LinkedDataProofException,
    ProofPurpose,
    derive,
    sign,
)
from ..ld_proofs.constants import CREDENTIALS_CONTEXT_V1_URL
from .models.credential import VerifiableCredentialSchema


async def create_presentation(
    *, credentials: List[dict], presentation_id: Optional[str] = None
) -> dict:
    """Create presentation and add the credentials to it.

    Will validates the structure off all credentials, but does
    not sign the presentation yet. Call sing_presentation to do this.

    Args:
        credentials (List[dict]): Credentials to add to the presentation
        presentation_id (str, optional): Id of the presentation. Defaults to None.

    Raises:
        LinkedDataProofException: When not all credentials have a valid structure

    Returns:
        dict: The unsigned presentation object

    """
    presentation = {
        "@context": [CREDENTIALS_CONTEXT_V1_URL],
        "type": ["VerifiablePresentation"],
    }

    # Validate structure of all credentials
    errors = VerifiableCredentialSchema().validate(credentials, many=True)
    if len(errors) > 0:
        raise LinkedDataProofException(
            f"Not all credentials have a valid structure: {errors}"
        )

    presentation["verifiableCredential"] = credentials

    if presentation_id:
        presentation["id"] = presentation_id

    return presentation


async def sign_presentation(
    *,
    presentation: dict,
    suite: LinkedDataProof,
    document_loader: DocumentLoaderMethod,
    purpose: Optional[ProofPurpose] = None,
    challenge: Optional[str] = None,
    domain: Optional[str] = None,
) -> dict:
    """Sign the presentation with the passed signature suite.

    Will set a default AuthenticationProofPurpose if no proof purpose is passed.

    Args:
        presentation (dict): The presentation to sign
        suite (LinkedDataProof): The signature suite to sign the presentation with
        document_loader (DocumentLoader): Document loader to use.
        purpose (ProofPurpose, optional): Purpose to use. Required if challenge is None
        challenge (str, optional): Challenge to use. Required if domain is None.
        domain (str, optional): Domain to use. Only used if purpose is None.

    Raises:
        LinkedDataProofException: When both purpose and challenge are not provided
            And when signing of the presentation fails

    Returns:
        dict: A verifiable presentation object

    """
    if not purpose and not challenge:
        raise LinkedDataProofException(
            'A "challenge" param is required when not providing a'
            ' "purpose" (for AuthenticationProofPurpose).'
        )
    if not purpose:
        purpose = AuthenticationProofPurpose(challenge=challenge, domain=domain)

    # TODO: validate structure of presentation

    return await sign(
        document=presentation,
        suite=suite,
        purpose=purpose,
        document_loader=document_loader,
    )


async def derive_credential(
    *,
    credential: dict,
    reveal_document: dict,
    suite: LinkedDataProof,
    document_loader: DocumentLoaderMethod,
) -> dict:
    """Derive new credential from the existing credential and the reveal document.

    All proofs matching the signature suite type will be replaced with a derived
    proof. Other proofs will be discarded.

    Args:
        credential (dict): The credential to derive the new credential from.
        reveal_document (dict): JSON-LD frame to select which attributes to include.
        suite (LinkedDataProof): The signature suite to use for derivation
        document_loader (DocumentLoader): The document loader to use.

    Returns:
        dict: The derived credential.

    """

    # Validate credential structure
    errors = VerifiableCredentialSchema().validate(credential)
    if len(errors) > 0:
        raise LinkedDataProofException(
            f"Unable to derive from credential with invalid structure: {errors}"
        )

    return await derive(
        document=credential,
        reveal_document=reveal_document,
        suite=suite,
        document_loader=document_loader,
    )



"""Helper classes and functions for message tests."""

from typing import Optional
from unittest import mock

from ......messaging.agent_message import AgentMessage, AgentMessageSchema
from .....didcomm_prefix import DIDCommPrefix


class MessageTest:
    """Base class for message tests."""

    TYPE: Optional[str] = None
    CLASS = AgentMessage
    SCHEMA = AgentMessageSchema
    VALUES: dict = {}

    def setUp(self):
        self.message = self.CLASS(**self.VALUES)

    def test_init(self):
        """Test initialization of message."""
        for key, value in self.VALUES.items():
            assert hasattr(self.message, key) and getattr(self.message, key) == value

    def test_type(self):
        """Test type matches expected."""
        assert self.message._type == DIDCommPrefix.qualify_current(self.TYPE)

    def test_deserialize(self):
        """Test deserialization of message."""
        with mock.patch.object(self.SCHEMA, "load") as message_schema_load:
            obj = {"obj": "obj"}

            message = self.CLASS.deserialize(obj)
            message_schema_load.assert_called_once_with(obj)

            assert message is message_schema_load.return_value

    def test_serialize(self):
        """Test serialization of message."""
        with mock.patch.object(self.SCHEMA, "dump") as message_schema_dump:
            message_dict = self.message.serialize()
            message_schema_dump.assert_called_once_with(self.message)

            assert message_dict is message_schema_dump.return_value

    def test_make_model(self):
        """Test serialize then deserialize results in instance of model."""
        message = self.CLASS(**self.VALUES)
        data = message.serialize()
        model_instance = self.CLASS.deserialize(data)
        assert isinstance(model_instance, self.CLASS)



from unittest import IsolatedAsyncioTestCase

from ......core.oob_processor import OobMessageProcessor
from ......indy.holder import IndyHolder
from ......indy.models.pres_preview import (
    IndyPresAttrSpec,
    IndyPresPredSpec,
    IndyPresPreview,
)
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......storage.error import StorageNotFoundError
from ......tests import mock
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from .....didcomm_prefix import DIDCommPrefix
from ...messages.presentation_proposal import PresentationProposal
from ...messages.presentation_request import PresentationRequest
from .. import presentation_request_handler as test_module

S_ID = "NcYxiDXkpYi6ov5FcYDi1e:2:vidya:1.0"
CD_ID = f"NcYxiDXkpYi6ov5FcYDi1e:3:CL:{S_ID}:tag1"
INDY_PROOF_REQ = {
    "name": "proof-req",
    "version": "1.0",
    "nonce": "12345",
    "requested_attributes": {
        "0_player_uuid": {
            "name": "player",
            "restrictions": [{"cred_def_id": CD_ID}],
        },
        "0_screencapture_uuid": {
            "name": "screenCapture",
            "restrictions": [{"cred_def_id": CD_ID}],
        },
    },
    "requested_predicates": {
        "0_highscore_GE_uuid": {
            "name": "highScore",
            "p_type": ">=",
            "p_value": 1000000,
            "restrictions": [{"cred_def_id": CD_ID}],
        }
    },
}
PRES_PREVIEW = IndyPresPreview(
    attributes=[
        IndyPresAttrSpec(name="player", cred_def_id=CD_ID, value="Richie Knucklez"),
        IndyPresAttrSpec(
            name="screenCapture",
            cred_def_id=CD_ID,
            mime_type="image/png",
            value="aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
        ),
    ],
    predicates=[
        IndyPresPredSpec(
            name="highScore", cred_def_id=CD_ID, predicate=">=", threshold=1000000
        )
    ],
)


class TestPresentationRequestHandler(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.request_context = RequestContext.test_context(await create_test_profile())
        self.mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        self.mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=mock.MagicMock()
        )
        self.request_context.injector.bind_instance(
            OobMessageProcessor, self.mock_oob_processor
        )

    async def test_called(self):
        self.request_context.connection_record = mock.MagicMock()
        self.request_context.connection_record.connection_id = "dummy"
        self.request_context.message_receipt = MessageReceipt()
        self.request_context.message = PresentationRequest()
        self.request_context.message.indy_proof_request = mock.MagicMock(
            return_value=INDY_PROOF_REQ
        )

        px_rec_instance = test_module.V10PresentationExchange(
            presentation_proposal_dict={
                "presentation_proposal": {
                    "@type": DIDCommPrefix.qualify_current(
                        "present-proof/1.0/presentation-preview"
                    ),
                    "attributes": [
                        {"name": "favourite", "cred_def_id": CD_ID, "value": "potato"},
                        {"name": "icon", "cred_def_id": CD_ID, "value": "cG90YXRv"},
                    ],
                    "predicates": [],
                }
            },
            auto_present=True,
        )

        with (
            mock.patch.object(
                test_module, "PresentationManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V10PresentationExchange", autospec=True
            ) as mock_pres_ex_cls,
        ):
            mock_pres_ex_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.receive_request = mock.CoroutineMock(
                return_value=mock.MagicMock()
            )
            mock_pres_mgr.return_value.receive_request.return_value.auto_present = False

            self.request_context.connection_ready = True
            handler = test_module.PresentationRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)

        mock_pres_mgr.return_value.receive_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        assert not responder.messages

    async def test_called_not_found(self):
        self.request_context.connection_record = mock.MagicMock()
        self.request_context.connection_record.connection_id = "dummy"
        self.request_context.message_receipt = MessageReceipt()
        self.request_context.message = PresentationRequest()
        self.request_context.message.indy_proof_request = mock.MagicMock(
            return_value=INDY_PROOF_REQ
        )

        px_rec_instance = test_module.V10PresentationExchange(
            presentation_proposal_dict={
                "presentation_proposal": {
                    "@type": DIDCommPrefix.qualify_current(
                        "present-proof/1.0/presentation-preview"
                    ),
                    "attributes": [
                        {"name": "favourite", "cred_def_id": CD_ID, "value": "potato"},
                        {"name": "icon", "cred_def_id": CD_ID, "value": "cG90YXRv"},
                    ],
                    "predicates": [],
                }
            },
            auto_present=True,
        )

        with (
            mock.patch.object(
                test_module, "PresentationManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V10PresentationExchange", autospec=True
            ) as mock_pres_ex_cls,
        ):
            mock_pres_ex_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                side_effect=StorageNotFoundError
            )
            mock_pres_ex_cls.return_value = px_rec_instance

            mock_pres_mgr.return_value.receive_request = mock.CoroutineMock(
                return_value=mock.MagicMock()
            )
            mock_pres_mgr.return_value.receive_request.return_value.auto_present = False

            self.request_context.connection_ready = True
            handler = test_module.PresentationRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)

        mock_pres_mgr.return_value.receive_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        assert not responder.messages

    async def test_called_auto_present(self):
        self.request_context.connection_record = mock.MagicMock()
        self.request_context.connection_record.connection_id = "dummy"
        self.request_context.message = PresentationRequest()
        self.request_context.message.indy_proof_request = mock.MagicMock(
            return_value={
                "name": "proof-request",
                "version": "1.0",
                "nonce": "1234567890",
                "requested_attributes": {
                    "0_favourite_uuid": {
                        "name": "favourite",
                        "restrictions": [
                            {
                                "cred_def_id": CD_ID,
                            }
                        ],
                    },
                    "1_icon_uuid": {
                        "name": "icon",
                        "restrictions": [
                            {
                                "cred_def_id": CD_ID,
                            }
                        ],
                    },
                },
                "requested_predicates": {},
            }
        )
        self.request_context.message_receipt = MessageReceipt()
        presentation_proposal = PresentationProposal(
            comment="Hello World", presentation_proposal=PRES_PREVIEW
        )
        px_rec_instance = test_module.V10PresentationExchange(
            presentation_proposal_dict=presentation_proposal,
            auto_present=True,
        )

        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=[{"cred_info": {"referent": "dummy"}}])
        )
        self.request_context.injector.bind_instance(IndyHolder, mock_holder)

        with (
            mock.patch.object(
                test_module, "PresentationManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V10PresentationExchange", autospec=True
            ) as mock_pres_ex_cls,
        ):
            mock_pres_ex_cls.return_value = px_rec_instance
            mock_pres_ex_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_presentation = mock.CoroutineMock(
                return_value=(px_rec_instance, "presentation_message")
            )
            self.request_context.connection_ready = True
            handler = test_module.PresentationRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_presentation.assert_called_once()

        mock_pres_mgr.return_value.receive_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "presentation_message"
        assert target == {}

    async def test_called_auto_present_x(self):
        self.request_context.connection_record = mock.MagicMock()
        self.request_context.connection_record.connection_id = "dummy"
        self.request_context.message = PresentationRequest()
        self.request_context.message.indy_proof_request = mock.MagicMock(
            return_value={
                "name": "proof-request",
                "version": "1.0",
                "nonce": "1234567890",
                "requested_attributes": {
                    "0_favourite_uuid": {
                        "name": "favourite",
                        "restrictions": [
                            {
                                "cred_def_id": CD_ID,
                            }
                        ],
                    },
                    "1_icon_uuid": {
                        "name": "icon",
                        "restrictions": [
                            {
                                "cred_def_id": CD_ID,
                            }
                        ],
                    },
                },
                "requested_predicates": {},
            }
        )
        self.request_context.message_receipt = MessageReceipt()
        presentation_proposal = PresentationProposal(
            comment="Hello World", presentation_proposal=PRES_PREVIEW
        )
        mock_px_rec = mock.MagicMock(
            presentation_proposal_dict=presentation_proposal,
            auto_present=True,
            save_error_state=mock.CoroutineMock(),
        )

        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=[{"cred_info": {"referent": "dummy"}}])
        )
        self.request_context.injector.bind_instance(IndyHolder, mock_holder)

        with (
            mock.patch.object(
                test_module, "PresentationManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V10PresentationExchange", autospec=True
            ) as mock_pres_ex_cls,
        ):
            mock_pres_ex_cls.return_value = mock_px_rec
            mock_pres_ex_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=mock_px_rec
            )
            mock_pres_mgr.return_value.receive_request = mock.CoroutineMock(
                return_value=mock_px_rec
            )

            mock_pres_mgr.return_value.create_presentation = mock.CoroutineMock(
                side_effect=test_module.IndyHolderError()
            )

            self.request_context.connection_ready = True
            handler = test_module.PresentationRequestHandler()
            responder = MockResponder()

            with mock.patch.object(
                handler._logger, "exception", mock.MagicMock()
            ) as mock_log_exc:
                await handler.handle(self.request_context, responder)
                mock_log_exc.assert_called_once()

    async def test_called_auto_present_no_preview(self):
        self.request_context.connection_record = mock.MagicMock()
        self.request_context.connection_record.connection_id = "dummy"
        self.request_context.message = PresentationRequest()
        self.request_context.message.indy_proof_request = mock.MagicMock(
            return_value={
                "name": "proof-request",
                "version": "1.0",
                "nonce": "1234567890",
                "requested_attributes": {
                    "0_favourite_uuid": {
                        "name": "favourite",
                        "restrictions": [
                            {
                                "cred_def_id": CD_ID,
                            }
                        ],
                    },
                    "1_icon_uuid": {
                        "name": "icon",
                        "restrictions": [
                            {
                                "cred_def_id": CD_ID,
                            }
                        ],
                    },
                },
                "requested_predicates": {},
            }
        )
        self.request_context.message_receipt = MessageReceipt()
        px_rec_instance = test_module.V10PresentationExchange(auto_present=True)

        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=[{"cred_info": {"referent": "dummy"}}])
        )
        self.request_context.injector.bind_instance(IndyHolder, mock_holder)

        with (
            mock.patch.object(
                test_module, "PresentationManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V10PresentationExchange", autospec=True
            ) as mock_pres_ex_cls,
        ):
            mock_pres_ex_cls.return_value = px_rec_instance
            mock_pres_ex_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_presentation = mock.CoroutineMock(
                return_value=(px_rec_instance, "presentation_message")
            )
            self.request_context.connection_ready = True
            handler = test_module.PresentationRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_presentation.assert_called_once()

        mock_pres_mgr.return_value.receive_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "presentation_message"
        assert target == {}

    async def test_called_auto_present_pred_no_match(self):
        self.request_context.connection_record = mock.MagicMock()
        self.request_context.connection_record.connection_id = "dummy"
        self.request_context.message = PresentationRequest()
        self.request_context.message.indy_proof_request = mock.MagicMock(
            return_value={
                "name": "proof-request",
                "version": "1.0",
                "nonce": "1234567890",
                "requested_attributes": {},
                "requested_predicates": {
                    "0_score_GE_uuid": {
                        "name": "score",
                        "p_type": ">=",
                        "p_value": 1000000,
                        "restrictions": [
                            {
                                "cred_def_id": CD_ID,
                            }
                        ],
                    }
                },
            }
        )
        self.request_context.message_receipt = MessageReceipt()
        px_rec_instance = test_module.V10PresentationExchange(auto_present=True)

        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=[])
        )
        self.request_context.injector.bind_instance(IndyHolder, mock_holder)

        with (
            mock.patch.object(
                test_module, "PresentationManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V10PresentationExchange", autospec=True
            ) as mock_pres_ex_cls,
        ):
            mock_pres_ex_cls.return_value = px_rec_instance
            mock_pres_ex_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_presentation = mock.CoroutineMock(
                return_value=(px_rec_instance, "presentation_message")
            )
            self.request_context.connection_ready = True
            handler = test_module.PresentationRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_presentation.assert_not_called()

        mock_pres_mgr.return_value.receive_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        assert not responder.messages

    async def test_called_auto_present_pred_single_match(self):
        self.request_context.connection_record = mock.MagicMock()
        self.request_context.connection_record.connection_id = "dummy"
        self.request_context.message = PresentationRequest()
        self.request_context.message.indy_proof_request = mock.MagicMock(
            return_value={
                "name": "proof-request",
                "version": "1.0",
                "nonce": "1234567890",
                "requested_attributes": {},
                "requested_predicates": {
                    "0_score_GE_uuid": {
                        "name": "score",
                        "p_type": ">=",
                        "p_value": 1000000,
                        "restrictions": [
                            {
                                "cred_def_id": CD_ID,
                            }
                        ],
                    }
                },
            }
        )
        self.request_context.message_receipt = MessageReceipt()
        px_rec_instance = test_module.V10PresentationExchange(auto_present=True)

        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=[{"cred_info": {"referent": "dummy-0"}}])
        )
        self.request_context.injector.bind_instance(IndyHolder, mock_holder)

        with (
            mock.patch.object(
                test_module, "PresentationManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V10PresentationExchange", autospec=True
            ) as mock_pres_ex_cls,
        ):
            mock_pres_ex_cls.return_value = px_rec_instance
            mock_pres_ex_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_presentation = mock.CoroutineMock(
                return_value=(px_rec_instance, "presentation_message")
            )
            self.request_context.connection_ready = True
            handler = test_module.PresentationRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_presentation.assert_called_once()

        mock_pres_mgr.return_value.receive_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "presentation_message"
        assert target == {}

    async def test_called_auto_present_pred_multi_match(self):
        self.request_context.connection_record = mock.MagicMock()
        self.request_context.connection_record.connection_id = "dummy"
        self.request_context.message = PresentationRequest()
        self.request_context.message.indy_proof_request = mock.MagicMock(
            return_value={
                "name": "proof-request",
                "version": "1.0",
                "nonce": "1234567890",
                "requested_attributes": {},
                "requested_predicates": {
                    "0_score_GE_uuid": {
                        "name": "score",
                        "p_type": ">=",
                        "p_value": 1000000,
                        "restrictions": [
                            {
                                "cred_def_id": CD_ID,
                            }
                        ],
                    }
                },
            }
        )
        self.request_context.message_receipt = MessageReceipt()
        px_rec_instance = test_module.V10PresentationExchange(auto_present=True)

        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(
                return_value=[
                    {"cred_info": {"referent": "dummy-0"}},
                    {"cred_info": {"referent": "dummy-1"}},
                ]
            )
        )
        self.request_context.injector.bind_instance(IndyHolder, mock_holder)

        with (
            mock.patch.object(
                test_module, "PresentationManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V10PresentationExchange", autospec=True
            ) as mock_pres_ex_cls,
        ):
            mock_pres_ex_cls.return_value = px_rec_instance
            mock_pres_ex_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_presentation = mock.CoroutineMock(
                return_value=(px_rec_instance, "presentation_message")
            )
            self.request_context.connection_ready = True
            handler = test_module.PresentationRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_presentation.assert_called_once()

        mock_pres_mgr.return_value.receive_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "presentation_message"
        assert target == {}

    async def test_called_auto_present_multi_cred_match_reft(self):
        self.request_context.connection_record = mock.MagicMock()
        self.request_context.connection_record.connection_id = "dummy"
        self.request_context.message = PresentationRequest()
        self.request_context.message.indy_proof_request = mock.MagicMock(
            return_value={
                "name": "proof-request",
                "version": "1.0",
                "nonce": "1234567890",
                "requested_attributes": {
                    "0_favourite_uuid": {
                        "name": "favourite",
                        "restrictions": [
                            {
                                "cred_def_id": CD_ID,
                            }
                        ],
                    },
                    "1_icon_uuid": {
                        "name": "icon",
                        "restrictions": [
                            {
                                "cred_def_id": CD_ID,
                            }
                        ],
                    },
                },
                "requested_predicates": {},
            }
        )
        self.request_context.message_receipt = MessageReceipt()
        px_rec_instance = test_module.V10PresentationExchange(
            presentation_proposal_dict={
                "presentation_proposal": {
                    "@type": DIDCommPrefix.qualify_current(
                        "present-proof/1.0/presentation-preview"
                    ),
                    "attributes": [
                        {"name": "favourite", "cred_def_id": CD_ID, "value": "potato"},
                        {"name": "icon", "cred_def_id": CD_ID, "value": "cG90YXRv"},
                    ],
                    "predicates": [],
                }
            },
            auto_present=True,
        )

        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(
                return_value=[
                    {
                        "cred_info": {
                            "referent": "dummy-0",
                            "cred_def_id": CD_ID,
                            "attrs": {
                                "ident": "zero",
                                "favourite": "potato",
                                "icon": "cG90YXRv",
                            },
                        }
                    },
                    {
                        "cred_info": {
                            "referent": "dummy-1",
                            "cred_def_id": CD_ID,
                            "attrs": {
                                "ident": "one",
                                "favourite": "spud",
                                "icon": "c3B1ZA==",
                            },
                        }
                    },
                    {
                        "cred_info": {
                            "referent": "dummy-2",
                            "cred_def_id": CD_ID,
                            "attrs": {
                                "ident": "two",
                                "favourite": "patate",
                                "icon": "cGF0YXRl",
                            },
                        }
                    },
                ]
            )
        )

        self.request_context.injector.bind_instance(IndyHolder, mock_holder)

        with (
            mock.patch.object(
                test_module, "PresentationManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V10PresentationExchange", autospec=True
            ) as mock_pres_ex_cls,
        ):
            mock_pres_ex_cls.return_value = px_rec_instance
            mock_pres_ex_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_presentation = mock.CoroutineMock(
                return_value=(px_rec_instance, "presentation_message")
            )
            self.request_context.connection_ready = True
            handler = test_module.PresentationRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_presentation.assert_called_once()

        mock_pres_mgr.return_value.receive_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "presentation_message"
        assert target == {}

    async def test_called_auto_present_bait_and_switch(self):
        self.request_context.connection_record = mock.MagicMock()
        self.request_context.connection_record.connection_id = "dummy"
        self.request_context.message = PresentationRequest()
        self.request_context.message.indy_proof_request = mock.MagicMock(
            return_value={
                "name": "proof-request",
                "version": "1.0",
                "nonce": "1234567890",
                "requested_attributes": {
                    "0_favourite_uuid": {
                        "name": "favourite",
                        "restrictions": [
                            {
                                "cred_def_id": CD_ID,
                            }
                        ],
                    }
                },
                "requested_predicates": {},
            }
        )
        self.request_context.message_receipt = MessageReceipt()
        px_rec_instance = test_module.V10PresentationExchange(
            presentation_proposal_dict={
                "presentation_proposal": {
                    "@type": DIDCommPrefix.qualify_current(
                        "present-proof/1.0/presentation-preview"
                    ),
                    "attributes": [
                        {"name": "favourite", "cred_def_id": CD_ID, "value": "potato"}
                    ],
                    "predicates": [],
                }
            },
            auto_present=True,
        )

        by_reft = mock.CoroutineMock(
            return_value=[
                {
                    "cred_info": {
                        "referent": "dummy-0",
                        "cred_def_id": CD_ID,
                        "attrs": {"ident": "zero", "favourite": "yam"},
                    }
                },
                {
                    "cred_info": {
                        "referent": "dummy-1",
                        "cred_def_id": CD_ID,
                        "attrs": {"ident": "one", "favourite": "turnip"},
                    }
                },
                {
                    "cred_info": {
                        "referent": "dummy-2",
                        "cred_def_id": CD_ID,
                        "attrs": {
                            "ident": "two",
                            "favourite": "the idea of a potato but not a potato",
                        },
                    }
                },
            ]
        )
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials_for_presentation_request_by_referent = by_reft
        self.request_context.injector.bind_instance(IndyHolder, mock_holder)

        with (
            mock.patch.object(
                test_module, "PresentationManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V10PresentationExchange", autospec=True
            ) as mock_pres_ex_cls,
        ):
            mock_pres_ex_cls.return_value = px_rec_instance
            mock_pres_ex_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_presentation = mock.CoroutineMock(
                return_value=(px_rec_instance, "presentation_message")
            )
            self.request_context.connection_ready = True
            handler = test_module.PresentationRequestHandler()
            responder = MockResponder()

            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_presentation.assert_not_called()

        mock_pres_mgr.return_value.receive_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        assert not responder.messages

    async def test_called_not_ready(self):
        self.request_context.message_receipt = MessageReceipt()
        self.request_context.connection_record = mock.MagicMock()

        with mock.patch.object(
            test_module, "PresentationManager", autospec=True
        ) as mock_pres_mgr:
            mock_pres_mgr.return_value.receive_request = mock.CoroutineMock()
            self.request_context.message = PresentationRequest()
            self.request_context.connection_ready = False
            handler = test_module.PresentationRequestHandler()
            responder = MockResponder()
            with self.assertRaises(test_module.HandlerException) as err:
                await handler.handle(self.request_context, responder)
            assert (
                err.exception.message
                == "Connection used for presentation request not ready"
            )

        assert not responder.messages

    async def test_no_conn_no_oob(self):
        self.request_context.message_receipt = MessageReceipt()
        self.mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        self.mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=None
        )
        self.request_context.injector.bind_instance(
            OobMessageProcessor, self.mock_oob_processor
        )
        self.request_context.message = PresentationRequest()
        handler = test_module.PresentationRequestHandler()
        responder = MockResponder()
        with self.assertRaises(test_module.HandlerException) as err:
            await handler.handle(self.request_context, responder)
        assert (
            err.exception.message
            == "No connection or associated connectionless exchange found for presentation request"
        )

        assert not responder.messages



from unittest import TestCase, mock

import pytest

from ......messaging.models.base import BaseModelError
from .....didcomm_prefix import DIDCommPrefix
from ...message_types import PRESENTATION_PROBLEM_REPORT, PROTOCOL_PACKAGE
from ..presentation_problem_report import PresentationProblemReport, ProblemReportReason


class TestPresentationProblemReport(TestCase):
    """Problem report tests."""

    def test_init_type(self):
        """Test initializer."""

        prob = PresentationProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ABANDONED.value,
            }
        )
        assert prob._type == DIDCommPrefix.qualify_current(PRESENTATION_PROBLEM_REPORT)

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.presentation_problem_report."
        "PresentationProblemReportSchema.load"
    )
    def test_deserialize(self, mock_load):
        """Test deserialization."""

        obj = PresentationProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ABANDONED.value,
            }
        )

        prob = PresentationProblemReport.deserialize(obj)
        mock_load.assert_called_once_with(obj)

        assert prob is mock_load.return_value

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.presentation_problem_report."
        "PresentationProblemReportSchema.dump"
    )
    def test_serialize(self, mock_dump):
        """Test serialization."""

        obj = PresentationProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ABANDONED.value,
            }
        )

        ser = obj.serialize()
        mock_dump.assert_called_once_with(obj)

        assert ser is mock_dump.return_value

    def test_make_model(self):
        """Test making model."""

        prob = PresentationProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ABANDONED.value,
            }
        )
        data = prob.serialize()
        model_instance = PresentationProblemReport.deserialize(data)
        assert isinstance(model_instance, PresentationProblemReport)

        prob = PresentationProblemReport()
        data = prob.serialize()
        with pytest.raises(BaseModelError):
            PresentationProblemReport.deserialize(data)



"""Base inbound transport class."""

from abc import ABC, abstractmethod
from collections import namedtuple
from typing import Awaitable, Callable, Optional

from ...core.profile import Profile
from ..error import TransportError
from ..wire_format import BaseWireFormat
from .session import InboundSession


class BaseInboundTransport(ABC):
    """Base inbound transport class."""

    def __init__(
        self,
        scheme: str,
        create_session: Callable,
        *,
        max_message_size: int = 0,
        is_external: bool = False,
        wire_format: Optional[BaseWireFormat] = None,
        root_profile: Optional[Profile] = None,
    ):
        """Initialize the inbound transport instance.

        Args:
            scheme (str): The transport scheme identifier.
            create_session (Callable): Method to create a new inbound session.
            max_message_size (int, optional): The maximum size of a message.
                Defaults to 0.
            is_external (bool, optional): Indicates if the transport is external.
                Defaults to False.
            wire_format (BaseWireFormat, optional): The wire format used for message
                serialization. Defaults to None.
            root_profile (Profile, optional): The root profile for the transport.
                Defaults to None.
        """
        self._create_session = create_session
        self._max_message_size = max_message_size
        self._scheme = scheme
        self.wire_format: BaseWireFormat = wire_format
        self.root_profile: Profile = root_profile
        self._is_external = is_external

    @property
    def max_message_size(self):
        """Accessor for this transport's max message size."""
        return self._max_message_size

    @property
    def scheme(self):
        """Accessor for this transport's scheme."""
        return self._scheme

    @property
    def is_external(self):
        """Accessor for this transport's is_external."""
        return self._is_external

    def create_session(
        self,
        *,
        accept_undelivered: bool = False,
        can_respond: bool = False,
        client_info: Optional[dict] = None,
        wire_format: Optional[BaseWireFormat] = None,
    ) -> Awaitable[InboundSession]:
        """Create a new inbound session.

        Args:
            accept_undelivered: Flag for accepting undelivered messages
            can_respond: Flag indicating that the transport can send responses
            client_info: Request-specific client information
            wire_format: Optionally override the session wire format
        """
        return self._create_session(
            accept_undelivered=accept_undelivered,
            can_respond=can_respond,
            client_info=client_info,
            wire_format=wire_format or self.wire_format,
            transport_type=self.scheme,
        )

    @abstractmethod
    async def start(self) -> None:
        """Start listening for on this transport."""

    @abstractmethod
    async def stop(self) -> None:
        """Stop listening for on this transport."""


class InboundTransportError(TransportError):
    """Generic inbound transport error."""


class InboundTransportRegistrationError(InboundTransportError):
    """Error in loading an inbound transport."""


class InboundTransportSetupError(InboundTransportError):
    """Setup error for an inbound transport."""


InboundTransportConfiguration = namedtuple(
    "InboundTransportConfiguration", "module host port"
)



"""Handle registration and publication of supported protocols."""

import logging
from dataclasses import dataclass
from typing import Any, Dict, Mapping, Optional, Sequence, Union

from ..config.injection_context import InjectionContext
from ..messaging.message_type import MessageType, MessageVersion, ProtocolIdentifier
from ..utils.classloader import ClassLoader, DeferLoad
from .error import ProtocolDefinitionValidationError, ProtocolMinorVersionNotSupported

LOGGER = logging.getLogger(__name__)


@dataclass
class VersionDefinition:
    """Version definition."""

    min: MessageVersion
    current: MessageVersion

    @classmethod
    def from_dict(cls, data: dict) -> "VersionDefinition":
        """Create a version definition from a dict."""
        return cls(
            min=MessageVersion(data["major_version"], data["minimum_minor_version"]),
            current=MessageVersion(data["major_version"], data["current_minor_version"]),
        )


@dataclass
class ProtocolDefinition:
    """Protocol metadata used to register and resolve message types."""

    ident: ProtocolIdentifier
    min: MessageVersion
    current: MessageVersion
    controller: Optional[str] = None

    @property
    def minor_versions_supported(self) -> bool:
        """Accessor for whether minor versions are supported."""
        return bool(self.current.minor >= 1 and self.current.minor >= self.min.minor)

    def __post_init__(self):
        """Post-init hook."""
        if self.min.major != self.current.major:
            raise ProtocolDefinitionValidationError(
                f"Major version mismatch: {self.min.major} != {self.current.major}"
            )
        if self.min.minor > self.current.minor:
            raise ProtocolDefinitionValidationError(
                f"Minimum minor version greater than current minor version: "
                f"{self.min.minor} > {self.current.minor}"
            )


class ProtocolRegistry:
    """Protocol registry for indexing message families."""

    def __init__(self):
        """Initialize a `ProtocolRegistry` instance."""

        self._definitions: Dict[str, ProtocolDefinition] = {}
        self._type_to_message_cls: Dict[str, Union[DeferLoad, type]] = {}

        # Mapping[protocol identifier, controller module path]
        self._controllers = {}

    @property
    def protocols(self) -> Sequence[str]:
        """Accessor for a list of all message protocols."""
        return [
            str(definition.ident.with_version((definition.min.major, minor)))
            for definition in self._definitions.values()
            for minor in range(definition.min.minor, definition.current.minor + 1)
        ]

    @property
    def message_types(self) -> Sequence[str]:
        """Accessor for a list of all message types."""
        return tuple(self._type_to_message_cls.keys())

    def protocols_matching_query(self, query: str) -> Sequence[str]:
        """Return a list of message protocols matching a query string."""
        all_types = self.protocols
        result = None

        if query == "*" or query is None:
            result = all_types
        elif query:
            if query.endswith("*"):
                match = query[:-1]
                result = tuple(k for k in all_types if k.startswith(match))
            elif query in all_types:
                result = (query,)
        return result or ()

    def register_message_types(
        self,
        typeset: Mapping[str, Union[str, type]],
        version_definition: Optional[Union[dict[str, Any], VersionDefinition]] = None,
    ):
        """Add new supported message types.

        Args:
            typeset (Mapping[str, Union[str, type]]): Mappings of message types to
                register.
            version_definition (Optional[Union[dict[str, Any], VersionDefinition]]):
                Optional version definition dict.

        """
        if version_definition is not None and isinstance(version_definition, dict):
            version_definition = VersionDefinition.from_dict(version_definition)

        definitions_to_add = {}
        type_to_message_cls_to_add = {}

        for message_type, message_cls in typeset.items():
            parsed = MessageType.from_str(message_type)
            protocol = ProtocolIdentifier.from_message_type(parsed)
            if protocol.stem in definitions_to_add:
                definition = definitions_to_add[protocol.stem]
            elif protocol.stem in self._definitions:
                definition = self._definitions[protocol.stem]
            else:
                if version_definition:
                    definition = ProtocolDefinition(
                        ident=protocol,
                        min=version_definition.min,
                        current=version_definition.current,
                    )
                else:
                    definition = ProtocolDefinition(
                        ident=protocol,
                        min=protocol.version,
                        current=protocol.version,
                    )

                definitions_to_add[protocol.stem] = definition

            if isinstance(message_cls, str):
                message_cls = DeferLoad(message_cls)

            type_to_message_cls_to_add[message_type] = message_cls

            if definition.minor_versions_supported:
                for minor_version in range(
                    definition.min.minor, definition.current.minor + 1
                ):
                    updated_type = parsed.with_version(
                        (parsed.version.major, minor_version)
                    )
                    type_to_message_cls_to_add[str(updated_type)] = message_cls

        self._type_to_message_cls.update(type_to_message_cls_to_add)
        self._definitions.update(definitions_to_add)

    def register_controllers(self, *controller_sets):
        """Add new controllers.

        Args:
            controller_sets: Mappings of message families to coroutines

        """
        for controlset in controller_sets:
            self._controllers.update(controlset)

    def resolve_message_class(
        self, message_type: str
    ) -> Optional[Union[DeferLoad, type]]:
        """Resolve a message_type to a message class.

        Given a message type identifier, this method
        returns the corresponding registered message class.

        Args:
            message_type: Message type to resolve

        Returns:
            The resolved message class

        """
        if (message_cls := self._type_to_message_cls.get(message_type)) is not None:
            return message_cls

        parsed = MessageType.from_str(message_type)
        protocol = ProtocolIdentifier.from_message_type(parsed)
        if definition := self._definitions.get(protocol.stem):
            if parsed.version.minor < definition.min.minor:
                raise ProtocolMinorVersionNotSupported(
                    f"Minimum supported minor version is {definition.min.minor}."
                    f" Received {parsed.version.minor}."
                )

            # This code will only be reached if the received minor version is greater
            # than our current supported version. All directly supported minor
            # versions would be returned previously.
            message_type = str(parsed.with_version(definition.current))

            if (message_cls := self._type_to_message_cls.get(message_type)) is not None:
                return message_cls

        return None

    async def prepare_disclosed(
        self, context: InjectionContext, protocols: Sequence[str]
    ):
        """Call controllers and return publicly supported message families and roles."""
        published = []
        for protocol in protocols:
            result: Dict[str, Any] = {"pid": protocol}
            if protocol in self._controllers:
                ctl_cls = self._controllers[protocol]
                if isinstance(ctl_cls, str):
                    ctl_cls = ClassLoader.load_class(ctl_cls)
                ctl_instance = ctl_cls(protocol)
                if hasattr(ctl_instance, "check_access"):
                    allowed = await ctl_instance.check_access(context)
                    if not allowed:
                        # remove from published
                        continue
                if hasattr(ctl_instance, "determine_roles"):
                    roles = await ctl_instance.determine_roles(context)
                    if roles:
                        result["roles"] = list(roles)
            published.append(result)
        return published

    def __repr__(self) -> str:
        """Return a string representation for this class."""
        return "<{}>".format(self.__class__.__name__)



from unittest import TestCase, mock

from .....didcomm_prefix import DIDCommPrefix
from ...message_types import CREDENTIAL_ACK, PROTOCOL_PACKAGE
from ..credential_ack import CredentialAck


class TestCredentialAck(TestCase):
    """Credential ack tests"""

    def test_init(self):
        """Test initializer"""
        credential_ack = CredentialAck()

    def test_type(self):
        """Test type"""
        credential_ack = CredentialAck()

        assert credential_ack._type == DIDCommPrefix.qualify_current(CREDENTIAL_ACK)

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.credential_ack.CredentialAckSchema.load")
    def test_deserialize(self, mock_credential_ack_schema_load):
        """
        Test deserialize
        """
        obj = CredentialAck()

        credential_ack = CredentialAck.deserialize(obj)
        mock_credential_ack_schema_load.assert_called_once_with(obj)

        assert credential_ack is mock_credential_ack_schema_load.return_value

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.credential_ack.CredentialAckSchema.dump")
    def test_serialize(self, mock_credential_ack_schema_dump):
        """
        Test serialization.
        """
        obj = CredentialAck()

        credential_ack_dict = obj.serialize()
        mock_credential_ack_schema_dump.assert_called_once_with(obj)

        assert credential_ack_dict is mock_credential_ack_schema_dump.return_value

    def test_make_model(self):
        """Test making model."""

        credential_ack = CredentialAck()
        data = credential_ack.serialize()
        model_instance = CredentialAck.deserialize(data)
        assert isinstance(model_instance, CredentialAck)



"""Verifiable Credential issuance methods."""

from typing import Optional

from ..ld_proofs import (
    CredentialIssuancePurpose,
    DocumentLoaderMethod,
    LinkedDataProof,
    LinkedDataProofException,
    ProofPurpose,
    sign,
)
from .models.credential import CredentialSchema


async def issue(
    *,
    credential: dict,
    suite: LinkedDataProof,
    document_loader: DocumentLoaderMethod,
    purpose: Optional[ProofPurpose] = None,
) -> dict:
    """Issue a verifiable credential.

    Takes the base credential document, verifies it, and adds
    a digital signature to it.

    Args:
        credential (dict): Base credential document.
        suite (LinkedDataProof): Signature suite to sign the credential with.
        document_loader (DocumentLoader): Document loader to use
        purpose (ProofPurpose, optional): A proof purpose instance that will match
            proofs to be verified and ensure they were created according to the
            appropriate purpose. Default to CredentialIssuancePurpose

    Raises:
        LinkedDataProofException: When the credential has an invalid structure
            OR signing fails

    Returns:
        dict: The signed verifiable credential

    """
    # Validate credential
    errors = CredentialSchema().validate(credential)
    if len(errors) > 0:
        raise LinkedDataProofException(f"Credential contains invalid structure: {errors}")

    # Set default proof purpose if not set
    if not purpose:
        purpose = CredentialIssuancePurpose()

    # Sign the credential with LD proof
    signed_credential = await sign(
        document=credential,
        suite=suite,
        purpose=purpose,
        document_loader=document_loader,
    )

    return signed_credential



SECURITY_V2 = {
    "@context": [
        {"@version": 1.1},
        "https://w3id.org/security/v1",
        {
            "AesKeyWrappingKey2019": "sec:AesKeyWrappingKey2019",
            "DeleteKeyOperation": "sec:DeleteKeyOperation",
            "DeriveSecretOperation": "sec:DeriveSecretOperation",
            "EcdsaSecp256k1Signature2019": "sec:EcdsaSecp256k1Signature2019",
            "EcdsaSecp256r1Signature2019": "sec:EcdsaSecp256r1Signature2019",
            "EcdsaSecp256k1VerificationKey2019": "sec:EcdsaSecp256k1VerificationKey2019",
            "EcdsaSecp256r1VerificationKey2019": "sec:EcdsaSecp256r1VerificationKey2019",
            "Ed25519Signature2018": "sec:Ed25519Signature2018",
            "Ed25519VerificationKey2018": "sec:Ed25519VerificationKey2018",
            "EquihashProof2018": "sec:EquihashProof2018",
            "ExportKeyOperation": "sec:ExportKeyOperation",
            "GenerateKeyOperation": "sec:GenerateKeyOperation",
            "KmsOperation": "sec:KmsOperation",
            "RevokeKeyOperation": "sec:RevokeKeyOperation",
            "RsaSignature2018": "sec:RsaSignature2018",
            "RsaVerificationKey2018": "sec:RsaVerificationKey2018",
            "Sha256HmacKey2019": "sec:Sha256HmacKey2019",
            "SignOperation": "sec:SignOperation",
            "UnwrapKeyOperation": "sec:UnwrapKeyOperation",
            "VerifyOperation": "sec:VerifyOperation",
            "WrapKeyOperation": "sec:WrapKeyOperation",
            "X25519KeyAgreementKey2019": "sec:X25519KeyAgreementKey2019",
            "allowedAction": "sec:allowedAction",
            "assertionMethod": {
                "@id": "sec:assertionMethod",
                "@type": "@id",
                "@container": "@set",
            },
            "authentication": {
                "@id": "sec:authenticationMethod",
                "@type": "@id",
                "@container": "@set",
            },
            "capability": {"@id": "sec:capability", "@type": "@id"},
            "capabilityAction": "sec:capabilityAction",
            "capabilityChain": {
                "@id": "sec:capabilityChain",
                "@type": "@id",
                "@container": "@list",
            },
            "capabilityDelegation": {
                "@id": "sec:capabilityDelegationMethod",
                "@type": "@id",
                "@container": "@set",
            },
            "capabilityInvocation": {
                "@id": "sec:capabilityInvocationMethod",
                "@type": "@id",
                "@container": "@set",
            },
            "caveat": {"@id": "sec:caveat", "@type": "@id", "@container": "@set"},
            "challenge": "sec:challenge",
            "ciphertext": "sec:ciphertext",
            "controller": {"@id": "sec:controller", "@type": "@id"},
            "delegator": {"@id": "sec:delegator", "@type": "@id"},
            "equihashParameterK": {
                "@id": "sec:equihashParameterK",
                "@type": "xsd:integer",
            },
            "equihashParameterN": {
                "@id": "sec:equihashParameterN",
                "@type": "xsd:integer",
            },
            "invocationTarget": {"@id": "sec:invocationTarget", "@type": "@id"},
            "invoker": {"@id": "sec:invoker", "@type": "@id"},
            "jws": "sec:jws",
            "keyAgreement": {
                "@id": "sec:keyAgreementMethod",
                "@type": "@id",
                "@container": "@set",
            },
            "kmsModule": {"@id": "sec:kmsModule"},
            "parentCapability": {"@id": "sec:parentCapability", "@type": "@id"},
            "plaintext": "sec:plaintext",
            "proof": {"@id": "sec:proof", "@type": "@id", "@container": "@graph"},
            "proofPurpose": {"@id": "sec:proofPurpose", "@type": "@vocab"},
            "proofValue": "sec:proofValue",
            "referenceId": "sec:referenceId",
            "unwrappedKey": "sec:unwrappedKey",
            "verificationMethod": {"@id": "sec:verificationMethod", "@type": "@id"},
            "verifyData": "sec:verifyData",
            "wrappedKey": "sec:wrappedKey",
        },
    ]
}



"""Presentation verification and validation result classes."""

from typing import List, Optional

from marshmallow import fields

from ...messaging.models.base import BaseModel, BaseModelSchema
from ...vc.ld_proofs.validation_result import DocumentVerificationResultSchema
from ..ld_proofs import DocumentVerificationResult


class PresentationVerificationResult(BaseModel):
    """Presentation verification result class."""

    class Meta:
        """PresentationVerificationResult metadata."""

        schema_class = "PresentationVerificationResultSchema"

    def __init__(
        self,
        *,
        verified: bool,
        presentation_result: Optional[DocumentVerificationResult] = None,
        credential_results: Optional[List[DocumentVerificationResult]] = None,
        errors: Optional[List[str]] = None,
    ) -> None:
        """Create new PresentationVerificationResult instance."""
        self.verified = verified
        self.presentation_result = presentation_result
        self.credential_results = credential_results
        self.errors = errors

    def __repr__(self) -> str:
        """Return a human readable representation of this class.

        Returns:
            A human readable string for this class

        """
        items = ("{}={}".format(k, repr(v)) for k, v in self.__dict__.items())
        return "<{}({})>".format(self.__class__.__name__, ", ".join(items))

    def __eq__(self, other: object) -> bool:
        """Comparison between presentation verification results."""
        if isinstance(other, PresentationVerificationResult):
            return (
                self.verified == other.verified
                and self.presentation_result == other.presentation_result
                # check credential results list
                and (
                    # both not present
                    (not self.credential_results and not other.credential_results)
                    # both list and matching
                    or (
                        isinstance(self.credential_results, list)
                        and isinstance(other.credential_results, list)
                        and all(
                            self_result == other_result
                            for (self_result, other_result) in zip(
                                self.credential_results, other.credential_results
                            )
                        )
                    )
                )
                # check error list
                and (
                    # both not present
                    (not self.errors and not other.errors)
                    # both list and matching
                    or (
                        isinstance(self.errors, list)
                        and isinstance(other.errors, list)
                        and all(
                            self_error == other_error
                            for (self_error, other_error) in zip(
                                self.errors, other.errors
                            )
                        )
                    )
                )
            )
        return False


class PresentationVerificationResultSchema(BaseModelSchema):
    """Presentation verification result schema."""

    class Meta:
        """PresentationVerificationResultSchema metadata."""

        model_class = PresentationVerificationResult

    verified = fields.Bool(required=True)
    presentation_result = fields.Nested(DocumentVerificationResultSchema)
    credential_results = fields.List(fields.Nested(DocumentVerificationResultSchema))
    errors = fields.List(fields.Str())
